/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                             reLOC 0.06-hanshin                             */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// cnf.cpp / 0.06-hanshin_029
/*----------------------------------------------------------------------------*/
//
// Dimacs CNF formula production tools.
//
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "compile.h"
#include "defs.h"
#include "reloc.h"
#include "cnf.h"
#include "statistics.h"

using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/
// sIndex

    bool sIndex::operator>=(const sIndex &index) const
    {
	return (index <= *this);
    }


    bool sIndex::operator>(const sIndex &index) const
    {
	return (index < *this);
    }


/*----------------------------------------------------------------------------*/
// sIntegerIndex

    sIntegerIndex::sIntegerIndex(int value)
	: m_value(value)
    {
	// nothing
    }


    sIntegerIndex::sIntegerIndex(const sIntegerIndex &integer_index)
	: m_value(integer_index.m_value)
    {
	// nothing
    }


    const sIntegerIndex& sIntegerIndex::operator=(const sIntegerIndex &integer_index)
    {
	m_value = integer_index.m_value;

	return *this;
    }


    sIndex* sIntegerIndex::clone(void) const
    {
	sIndex *index = new sIntegerIndex(*this);
	return index;
    }


    int sIntegerIndex::get_Value(void) const
    {
	return m_value;
    }


    void sIntegerIndex::set_Value(int value)
    {
	m_value = value;
    }


    void sIntegerIndex::increment(void)
    {
	++m_value;
    }


    void sIntegerIndex::decrement(void)
    {
	--m_value;
    }


    void sIntegerIndex::increment(int distance)
    {
	m_value += distance;
    }


    void sIntegerIndex::decrement(int distance)
    {
	m_value -= distance;
    }


    bool sIntegerIndex::operator==(const sIndex &index) const
    {
	return (m_value == ((sIntegerIndex&)index).m_value);
    }


    bool sIntegerIndex::operator!=(const sIndex &index) const
    {
	return (m_value != ((sIntegerIndex&)index).m_value);
    }


    bool sIntegerIndex::operator<=(const sIndex &index) const
    {
	return (m_value <= ((sIntegerIndex&)index).m_value);
    }


    bool sIntegerIndex::operator<(const sIndex &index) const
    {
	return (m_value < ((sIntegerIndex&)index).m_value);
    }


    int sIntegerIndex::operator-(const sIndex &index) const
    {
	return (m_value - ((sIntegerIndex&)index).m_value);
    }


    sString sIntegerIndex::to_String(void) const
    {
	return sUInt_32_to_String(m_value);
    }


    void sIntegerIndex::to_Screen(void) const
    {
	printf("Integer index (value = %d)\n", m_value);
    }


/*----------------------------------------------------------------------------*/
// sIndex_iterator

    sIndex_iterator::sIndex_iterator(const sIndex &index)
    {
	m_index = index.clone();
    }


    sIndex_iterator::sIndex_iterator(const sIndex_iterator &index_iterator)
    {
	m_index = index_iterator.m_index->clone();
    }


    const sIndex_iterator& sIndex_iterator::operator=(const sIndex_iterator &index_iterator)
    {
	delete m_index;
	m_index = index_iterator.m_index->clone();
	return *this;
    }


    sIndex_iterator::~sIndex_iterator()
    {
	delete m_index;
    }


    sIndex_iterator& sIndex_iterator::operator++()
    {
	m_index->increment();
	return *this;
    }


    sIndex_iterator& sIndex_iterator::operator--()
    {
	m_index->decrement();
	return *this;
    }


    sIndex_iterator& sIndex_iterator::operator+=(int distance)
    {
	m_index->increment(distance);
	return *this;
    }


    sIndex_iterator& sIndex_iterator::operator-=(int distance)
    {
	m_index->decrement(distance);
	return *this;
    }
    

    bool sIndex_iterator::operator==(const sIndex_iterator &iterator) const
    {
	return (*m_index == *iterator);
    }


    bool sIndex_iterator::operator!=(const sIndex_iterator &iterator) const
    {
	return (*m_index != *iterator);
    }


    const sIndex& sIndex_iterator::operator*() const
    {
	return *m_index;
    }


/*----------------------------------------------------------------------------*/
// sIndexScope

    sIndexScope::sIndexScope()
    {
	// nothing
    }


    sIndexScope::sIndexScope(const sIndexScope &sUNUSED(scope))
    {
	// nothing
    }


    const sIndexScope& sIndexScope::operator=(const sIndexScope &sUNUSED(scope))
    {
	return *this;
    }
    

/*----------------------------------------------------------------------------*/
// sIntegerScope

    sIntegerScope::sIntegerScope()
	: m_min(0)
	, m_max(0)
    {
	// nothing
    }

    sIntegerScope::sIntegerScope(int min, int max)
	: m_min(min)
	, m_max(max)
    {
	// nothing
    }
    
    
    sIntegerScope::sIntegerScope(const sIntegerScope &integer_scope)
	: sIndexScope(integer_scope)
	, m_min(integer_scope.m_min)
	, m_max(integer_scope.m_max)
    {
	// nothing
    }

    
    const sIntegerScope& sIntegerScope::operator=(const sIntegerScope &integer_scope)
    {
	sIndexScope::operator=(integer_scope);
	m_min = integer_scope.m_min;
	m_max = integer_scope.m_max;

	return *this;
    }


    int sIntegerScope::get_Size(void) const
    {
	return m_max - m_min + 1;
    }

    
    sIndexScope* sIntegerScope::clone(void) const
    {
	sIndexScope *index_scope = new sIntegerScope(*this);
	return index_scope;
    }


    sIndex_iterator sIntegerScope::get_Begin(void) const
    {
	return sIndex_iterator(sIntegerIndex(m_min));
    }


    sIndex_iterator sIntegerScope::get_End(void) const
    {
	return sIndex_iterator(sIntegerIndex(m_max + 1));
    }


/*----------------------------------------------------------------------------*/
// sVariableStore_CNF

    sVariableStore_CNF::sVariableStore_CNF()
	: m_last_variable(1)
    {
	// nothing
    }


    sVariableStore_CNF::sVariableStore_CNF(const sVariableStore_CNF &variable_store)
	: m_last_variable(variable_store.m_last_variable)
    {
	// nothing
    }


    const sVariableStore_CNF& sVariableStore_CNF::operator=(const sVariableStore_CNF &variable_store)
    {
	m_last_variable = variable_store.m_last_variable;

	return *this;
    }

    int sVariableStore_CNF::get_Last_CNFVariable(void) const
    {
	return m_last_variable;
    }


    void sVariableStore_CNF::alloc_CNFVariables(int N_Variables)
    {
	m_last_variable += N_Variables;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Variables += N_Variables;
	}
	#endif
    }


/*----------------------------------------------------------------------------*/
// sSpecifiedIdentifier

    sSpecifiedIdentifier::sSpecifiedIdentifier()
	: m_indexable_identifier(NULL)
    {
    }


    sSpecifiedIdentifier::sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier)
	: m_indexable_identifier(indexable_identifier)
    {
	// nothing
    }


    sSpecifiedIdentifier::sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index)
	: m_indexable_identifier(indexable_identifier)
    {
	m_scope_Indexes.push_back(index.clone());
    }


    sSpecifiedIdentifier::sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index_1, const sIndex &index_2)
	: m_indexable_identifier(indexable_identifier)
    {
	m_scope_Indexes.push_back(index_1.clone());
	m_scope_Indexes.push_back(index_2.clone());
    }


    sSpecifiedIdentifier::sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const sIndex &index_1, const sIndex &index_2, const sIndex &index_3)
	: m_indexable_identifier(indexable_identifier)
    {
	m_scope_Indexes.push_back(index_1.clone());
	m_scope_Indexes.push_back(index_2.clone());
	m_scope_Indexes.push_back(index_3.clone());
    }


    sSpecifiedIdentifier::sSpecifiedIdentifier(const sIndexableIdentifier *indexable_identifier, const Indexes_vector &scope_Indexes)
	: m_indexable_identifier(indexable_identifier)
    {
	for (Indexes_vector::const_iterator index = scope_Indexes.begin(); index != scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}
    }
    

    sSpecifiedIdentifier::sSpecifiedIdentifier(const sSpecifiedIdentifier &specified_identifier)
	: m_indexable_identifier(specified_identifier.m_indexable_identifier)
    {
	for (Indexes_vector::const_iterator index = specified_identifier.m_scope_Indexes.begin(); index != specified_identifier.m_scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}
    }


    const sSpecifiedIdentifier& sSpecifiedIdentifier::operator=(const sSpecifiedIdentifier &specified_identifier)
    {
	m_indexable_identifier = specified_identifier.m_indexable_identifier;

	for (Indexes_vector::const_iterator index = m_scope_Indexes.begin(); index != m_scope_Indexes.end(); ++index)
	{
	    delete *index;
	}

	m_scope_Indexes.clear();

	for (Indexes_vector::const_iterator index = specified_identifier.m_scope_Indexes.begin(); index != specified_identifier.m_scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}

	return *this;
    }


    sSpecifiedIdentifier::~sSpecifiedIdentifier()
    {
	for (Indexes_vector::const_iterator index = m_scope_Indexes.begin(); index != m_scope_Indexes.end(); ++index)
	{
	    delete *index;
	}
    }
    

    sString sSpecifiedIdentifier::calc_String(void) const
    {
	switch(m_scope_Indexes.size())
	{
	case 0:
	{
	    return m_indexable_identifier->calc_String();
	    break;
	}
	case 1:
	{
	    return m_indexable_identifier->calc_String(*m_scope_Indexes[0]);
	    break;
	}
	case 2:
	{
	    return m_indexable_identifier->calc_String(*m_scope_Indexes[0], *m_scope_Indexes[1]);
	    break;
	}
	case 3:
	{
	    return m_indexable_identifier->calc_String(*m_scope_Indexes[0], *m_scope_Indexes[1], *m_scope_Indexes[2]);
	    break;
	}
	default:
	{
	    return m_indexable_identifier->calc_String(m_scope_Indexes);
	    break;
	}
	}
    }


    int sSpecifiedIdentifier::calc_CNF(void) const
    {
	switch(m_scope_Indexes.size())
	{
	case 0:
	{
	    return m_indexable_identifier->calc_CNF();
	    break;
	}
	case 1:
	{
	    return m_indexable_identifier->calc_CNF(*m_scope_Indexes[0]);
	    break;
	}
	case 2:
	{
	    return m_indexable_identifier->calc_CNF(*m_scope_Indexes[0], *m_scope_Indexes[1]);
	    break;
	}
	case 3:
	{
	    return m_indexable_identifier->calc_CNF(*m_scope_Indexes[0], *m_scope_Indexes[1], *m_scope_Indexes[2]);
	    break;
	}
	default:
	{
	    return m_indexable_identifier->calc_CNF(m_scope_Indexes);
	    break;
	}
	}
    }
    

/*----------------------------------------------------------------------------*/
// sIndexableIdentifier

    sIndexableIdentifier::sIndexableIdentifier()
	: m_variable_store(NULL)
    {
	// nothing
    }

    sIndexableIdentifier::sIndexableIdentifier(sVariableStore_CNF *variable_store, const sString &base_name)
	: m_variable_store(variable_store)
	, m_base_name(base_name)
    {
	m_first_cnf_variable = variable_store->get_Last_CNFVariable();
	m_N_cnf_Variables = 1;
	variable_store->alloc_CNFVariables(1);
    }


    sIndexableIdentifier::sIndexableIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, const sIndexScope &scope)
	: m_variable_store(variable_store)
	, m_base_name(base_name)
    {
	m_first_cnf_variable = variable_store->get_Last_CNFVariable();
	m_N_cnf_Variables = scope.get_Size();
	variable_store->alloc_CNFVariables(m_N_cnf_Variables);

	m_index_Scopes.push_back(scope.clone());
    }


    sIndexableIdentifier::sIndexableIdentifier(sVariableStore_CNF *variable_store,
					       const sString      &base_name,
					       const sIndexScope  &scope_1,
					       const sIndexScope  &scope_2)
	: m_variable_store(variable_store)
	, m_base_name(base_name)
    {
	m_first_cnf_variable = variable_store->get_Last_CNFVariable();
	m_N_cnf_Variables = scope_1.get_Size() * scope_2.get_Size();
	variable_store->alloc_CNFVariables(m_N_cnf_Variables);

	m_index_Scopes.push_back(scope_1.clone());
	m_index_Scopes.push_back(scope_2.clone());
    }


    sIndexableIdentifier::sIndexableIdentifier(sVariableStore_CNF *variable_store,
					       const sString      &base_name,
					       const sIndexScope  &scope_1,
					       const sIndexScope  &scope_2,
					       const sIndexScope  &scope_3)
	: m_variable_store(variable_store)
	, m_base_name(base_name)
    {
	m_first_cnf_variable = variable_store->get_Last_CNFVariable();
	m_N_cnf_Variables = scope_1.get_Size() * scope_2.get_Size() * scope_3.get_Size();
	variable_store->alloc_CNFVariables(m_N_cnf_Variables);

	m_index_Scopes.push_back(scope_1.clone());
	m_index_Scopes.push_back(scope_2.clone());
	m_index_Scopes.push_back(scope_3.clone());
    }


    sIndexableIdentifier::sIndexableIdentifier(sVariableStore_CNF       *variable_store,
					       const sString             &base_name,
					       const IndexScopes_vector &index_Scopes)
	: m_variable_store(variable_store)
	, m_base_name(base_name)
    {
	m_first_cnf_variable = variable_store->get_Last_CNFVariable();
	m_N_cnf_Variables = 1;

	for (IndexScopes_vector::const_iterator scope = index_Scopes.begin(); scope != index_Scopes.end(); ++scope)
	{
	    m_N_cnf_Variables *= (*scope)->get_Size();
	    sIndexScope *index_scope = (*scope)->clone();
	    m_index_Scopes.push_back(index_scope);
	}
	variable_store->alloc_CNFVariables(m_N_cnf_Variables);
    }


    sIndexableIdentifier::sIndexableIdentifier(const sIndexableIdentifier &indexable_identifier)
	: m_variable_store(indexable_identifier.m_variable_store)
	, m_base_name(indexable_identifier.m_base_name)
	, m_first_cnf_variable(indexable_identifier.m_first_cnf_variable)
	, m_N_cnf_Variables(indexable_identifier.m_N_cnf_Variables)
	  
    {
	for (IndexScopes_vector::const_iterator scope = indexable_identifier.m_index_Scopes.begin(); scope != indexable_identifier.m_index_Scopes.end(); ++scope)
	{
	    sIndexScope *index_scope = (*scope)->clone();
	    m_index_Scopes.push_back(index_scope);
	}	    
    }


    const sIndexableIdentifier& sIndexableIdentifier::operator=(const sIndexableIdentifier &indexable_identifier)
    {
	m_variable_store = indexable_identifier.m_variable_store;
	m_base_name = indexable_identifier.m_base_name;

	m_first_cnf_variable = indexable_identifier.m_first_cnf_variable;
	m_N_cnf_Variables = indexable_identifier.m_N_cnf_Variables;

	m_first_cnf_variable = indexable_identifier.m_first_cnf_variable;
	m_N_cnf_Variables = indexable_identifier.m_N_cnf_Variables;

	for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope)
	{
	    delete *scope;
	}

	m_index_Scopes.clear();

	for (IndexScopes_vector::const_iterator scope = indexable_identifier.m_index_Scopes.begin(); scope != indexable_identifier.m_index_Scopes.end(); ++scope)
	{
	    sIndexScope *index_scope = (*scope)->clone();
	    m_index_Scopes.push_back(index_scope);
	}	    
	
	return *this;
    }


    sIndexableIdentifier::~sIndexableIdentifier()
    {
	for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope)
	{
	    delete *scope;
	}
    }


    sString sIndexableIdentifier::calc_String(void) const
    {
	return m_base_name;
    }


    sString sIndexableIdentifier::calc_String(const sIndex &index) const
    {
	sASSERT(index >= *m_index_Scopes[0]->get_Begin() && index < *m_index_Scopes[0]->get_End());

	return m_base_name + "_" + index.to_String();
    }


    sString sIndexableIdentifier::calc_String(const sIndex &index_1, const sIndex &index_2) const
    {
	sASSERT(index_1 >= *m_index_Scopes[0]->get_Begin() && index_1 < *m_index_Scopes[0]->get_End());
	sASSERT(index_2 >= *m_index_Scopes[1]->get_Begin() && index_2 < *m_index_Scopes[1]->get_End());

	return m_base_name + "_" + index_1.to_String() + "^" + index_2.to_String();
    }


    sString sIndexableIdentifier::calc_String(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3) const
    {
	sASSERT(index_1 >= *m_index_Scopes[0]->get_Begin() && index_1 < *m_index_Scopes[0]->get_End());
	sASSERT(index_2 >= *m_index_Scopes[1]->get_Begin() && index_2 < *m_index_Scopes[1]->get_End());
	sASSERT(index_3 >= *m_index_Scopes[2]->get_Begin() && index_3 < *m_index_Scopes[2]->get_End());

	return m_base_name + "_" + index_1.to_String() + "^" + index_2.to_String() + "/" + index_3.to_String();;
    }


    sString sIndexableIdentifier::calc_String(const Indexes_vector &scope_Indexes) const
    {
	sString output = m_base_name;

	Indexes_vector::const_iterator index = scope_Indexes.begin();

	for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope, ++index)
	{
	    sASSERT(**index >= *(*scope)->get_Begin() && **index < *(*scope)->get_End());
	    output += "." + (*index)->to_String();
	}

	return output;
    }


    int sIndexableIdentifier::calc_CNF(void) const
    {
	return m_first_cnf_variable;
    }


    int sIndexableIdentifier::calc_CNF(const sIndex &index) const
    {
	sASSERT(index >= *m_index_Scopes[0]->get_Begin() && index < *m_index_Scopes[0]->get_End());

	return m_first_cnf_variable + (index - *m_index_Scopes[0]->get_Begin());
    }


    int sIndexableIdentifier::calc_CNF(const sIndex &index_1, const sIndex &index_2) const
    {
	sASSERT(index_1 >= *m_index_Scopes[0]->get_Begin() && index_1 < *m_index_Scopes[0]->get_End());
	sASSERT(index_2 >= *m_index_Scopes[1]->get_Begin() && index_2 < *m_index_Scopes[1]->get_End());

	return m_first_cnf_variable + (index_1 - *m_index_Scopes[0]->get_Begin()) * m_index_Scopes[1]->get_Size() + (index_2 - *m_index_Scopes[1]->get_Begin());
    }


    int sIndexableIdentifier::calc_CNF(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3) const
    {
	sASSERT(index_1 >= *m_index_Scopes[0]->get_Begin() && index_1 < *m_index_Scopes[0]->get_End());
	sASSERT(index_2 >= *m_index_Scopes[1]->get_Begin() && index_2 < *m_index_Scopes[1]->get_End());
	sASSERT(index_3 >= *m_index_Scopes[2]->get_Begin() && index_3 < *m_index_Scopes[2]->get_End());

	return m_first_cnf_variable + (index_1 - *m_index_Scopes[0]->get_Begin()) * m_index_Scopes[1]->get_Size() * m_index_Scopes[2]->get_Size()
                         	    + (index_2 - *m_index_Scopes[1]->get_Begin()) * m_index_Scopes[2]->get_Size()
                                    + (index_3 - *m_index_Scopes[2]->get_Begin());
    }


    int sIndexableIdentifier::calc_CNF(const Indexes_vector &scope_Indexes) const
    {
	int offset = m_N_cnf_Variables;
	int cnf_var = m_first_cnf_variable;

	Indexes_vector::const_iterator index = scope_Indexes.begin();

	for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope, ++index)
	{
	    sASSERT(**index >= *(*scope)->get_Begin() && **index < *(*scope)->get_End());
	    offset /= (*scope)->get_Size();
	    cnf_var += (**index - *(*scope)->get_Begin()) * offset; 
	}

	return cnf_var;
    }


    sSpecifiedIdentifier sIndexableIdentifier::translate_CNFVariable(int cnf_variable) const
    {
	sASSERT(m_variable_store != NULL);

	if (cnf_variable >= m_first_cnf_variable && cnf_variable < m_first_cnf_variable + m_N_cnf_Variables)
	{
	    int offset = m_N_cnf_Variables;
	    cnf_variable -= m_first_cnf_variable;
	    Indexes_vector identifier_Indexes;

	    for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope)
	    {
		offset /= (*scope)->get_Size();
		sIndex_iterator index_iter = (*scope)->get_Begin();
		index_iter += (cnf_variable / offset);

		identifier_Indexes.push_back((*index_iter).clone());
		cnf_variable %= offset;
	    }	    
	    sSpecifiedIdentifier specified_identifier(this, identifier_Indexes);

	    for (Indexes_vector::iterator index = identifier_Indexes.begin(); index != identifier_Indexes.end(); ++index)
	    {
		delete *index;
	    }
	    return specified_identifier;
	}
	else
	{
	    return sSpecifiedIdentifier();
	}
    }


    void sIndexableIdentifier::to_Screen(const sString &indent) const
    {
	printf("%sIndexable identifier (type = ", indent.c_str());

	switch (get_Arity())
	{
	case 0:
	{
	    printf("0-ary) [\n");
	    break;
	}
	case 1:
	{
	    printf("unary) [\n");
	    break;
	}
	case 2:
	{
	    printf("binary) [\n");
	    break;
	}
	case 3:
	{
	    printf("ternary) [\n");
	    break;
	}
	default:
	{
	    printf("%d-ary) [\n", get_Arity());
	    break;
	}
	}

	switch (get_Arity())
	{
	case 0:
	{
	    printf("%s%s%s->%d\n", sRELOC_INDENT.c_str(), indent.c_str(), m_base_name.c_str(), m_first_cnf_variable);
	    break;
	}
	case 1:
	{
	    printf("%s%s", sRELOC_INDENT.c_str(), indent.c_str());
		
	    sIndex_iterator end = m_index_Scopes[0]->get_End();
	    for (sIndex_iterator index = m_index_Scopes[0]->get_Begin(); index != end; ++index)
	    {		
		sString identifier = calc_String(*index);
		int cnf_index =  calc_CNF(*index);
		printf("%s -> %d\t", identifier.c_str(), cnf_index);
	    }
	    printf("\n");
	    break;
	}
	case 2:
	{
	    sIndex_iterator end_1 = m_index_Scopes[0]->get_End();
	    for (sIndex_iterator index_1 = m_index_Scopes[0]->get_Begin(); index_1 != end_1; ++index_1)
	    {
		printf("%s%s", sRELOC_INDENT.c_str(), indent.c_str());

		sIndex_iterator end_2 = m_index_Scopes[1]->get_End();
		for (sIndex_iterator index_2 = m_index_Scopes[1]->get_Begin(); index_2 != end_2; ++index_2)
		{
		    sString identifier = calc_String(*index_1, *index_2);
		    int cnf_index =  calc_CNF(*index_1, *index_2);
		    printf("%s -> %d\t", identifier.c_str(), cnf_index);
		}
		printf("\n");
	    }
	    break;
	}
	case 3:
	{
	    sIndex_iterator end_1 = m_index_Scopes[0]->get_End();
	    sIndex_iterator pre_end_1 = --m_index_Scopes[0]->get_End();
	    for (sIndex_iterator index_1 = m_index_Scopes[0]->get_Begin(); index_1 != end_1; ++index_1)
	    {
		sIndex_iterator end_2 = m_index_Scopes[1]->get_End();
		for (sIndex_iterator index_2 = m_index_Scopes[1]->get_Begin(); index_2 != end_2; ++index_2)
		{
		    printf("%s%s", sRELOC_INDENT.c_str(), indent.c_str());

		    sIndex_iterator end_3 = m_index_Scopes[2]->get_End();
		    for (sIndex_iterator index_3 = m_index_Scopes[2]->get_Begin(); index_3 != end_3; ++index_3)
		    {
			sString identifier = calc_String(*index_1, *index_2, *index_3);
			int cnf_index =  calc_CNF(*index_1, *index_2, *index_3);
			printf("%s -> %d\t", identifier.c_str(), cnf_index);
		    }
		    printf("\n");
		}
		if (index_1 != pre_end_1)
		{
		    printf("\n");
		}
	    }
	    break;
	}
	default:
	{
	    Indexes_vector Indexes;
	    to_Screen_identifiers(Indexes, indent);
	    printf("\n");
	}
	}

	printf("%s]\n", indent.c_str());
    }


/*----------------------------------------------------------------------------*/

    void sIndexableIdentifier::to_Screen_identifiers(const Indexes_vector &scope_Indexes, const sString &indent) const
    {
	if (scope_Indexes.size() == m_index_Scopes.size())
	{
	    sString output = m_base_name;
	    Indexes_vector::const_iterator index = scope_Indexes.begin();

	    for (IndexScopes_vector::const_iterator scope = m_index_Scopes.begin(); scope != m_index_Scopes.end(); ++scope, ++index)
	    {
		output += "." + (*index)->to_String();
	    }

	    printf("%s -> %d\t", output.c_str(), calc_CNF(scope_Indexes));
	}
	else
	{
	    if (m_index_Scopes.size() - scope_Indexes.size() == 1)
	    {
		printf("%s%s", indent.c_str(), sRELOC_INDENT.c_str());
	    }
	    const sIndexScope *scope = m_index_Scopes[scope_Indexes.size()];

	    sIndex_iterator end = scope->get_End();
	    for (sIndex_iterator index = scope->get_Begin(); index != end; ++index)
	    {
		Indexes_vector scope_Indexes_rec = scope_Indexes;
		scope_Indexes_rec.push_back(&(*index));
		to_Screen_identifiers(scope_Indexes_rec, indent);
	    }

	    bool line = false;
	    int last_index = scope_Indexes.size() - 1;

	    for (int i = 0; i <= last_index ; ++i)
	    {
		if (*scope_Indexes[i] != *(--m_index_Scopes[i]->get_End()))
		{
		    line = true;
		    break;
		}
	    }	
	    if (line)
	    {
		printf("\n");
	    }
	}
    }


/*----------------------------------------------------------------------------*/
// sSpecifiedIStatedentifier

    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier()
	: m_state_identifier(NULL)
    {
	// nothong
    }


    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier)
	: m_state_identifier(state_identifier)
    {
	// nothing
    }


    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index)
	: m_state_identifier(state_identifier)
    {
	m_scope_Indexes.push_back(index.clone());
    }


    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index_1, const sIndex &index_2)
	: m_state_identifier(state_identifier)
    {
	m_scope_Indexes.push_back(index_1.clone());
	m_scope_Indexes.push_back(index_2.clone());
    }


    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const sIndex &index_1, const sIndex &index_2, const sIndex &index_3)
	: m_state_identifier(state_identifier)
    {
	m_scope_Indexes.push_back(index_1.clone());
	m_scope_Indexes.push_back(index_2.clone());
	m_scope_Indexes.push_back(index_3.clone());
    }


    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sIndexableStateIdentifier *state_identifier, const Indexes_vector &scope_Indexes)
	: m_state_identifier(state_identifier)
    {
	for (Indexes_vector::const_iterator index = scope_Indexes.begin(); index != scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}
    }
    

    sSpecifiedStateIdentifier::sSpecifiedStateIdentifier(const sSpecifiedStateIdentifier &specified_identifier)
	: m_state_identifier(specified_identifier.m_state_identifier)
    {
	for (Indexes_vector::const_iterator index = specified_identifier.m_scope_Indexes.begin(); index != specified_identifier.m_scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}
    }


    const sSpecifiedStateIdentifier& sSpecifiedStateIdentifier::operator=(const sSpecifiedStateIdentifier &specified_identifier)
    {
	m_state_identifier = specified_identifier.m_state_identifier;

	for (Indexes_vector::const_iterator index = m_scope_Indexes.begin(); index != m_scope_Indexes.end(); ++index)
	{
	    delete *index;
	}

	m_scope_Indexes.clear();

	for (Indexes_vector::const_iterator index = specified_identifier.m_scope_Indexes.begin(); index != specified_identifier.m_scope_Indexes.end(); ++index)
	{
	    m_scope_Indexes.push_back((*index)->clone());
	}

	return *this;
    }


    sSpecifiedStateIdentifier::~sSpecifiedStateIdentifier()
    {
	for (Indexes_vector::const_iterator index = m_scope_Indexes.begin(); index != m_scope_Indexes.end(); ++index)
	{
	    delete *index;
	}
    }
    

    sString sSpecifiedStateIdentifier::calc_String(int bit) const
    {
	switch(m_scope_Indexes.size())
	{
	case 0:
	{
	    return m_state_identifier->calc_String(bit);
	    break;
	}
	case 1:
	{
	    return m_state_identifier->calc_String(*m_scope_Indexes[0], bit);
	    break;
	}
	case 2:
	{
	    return m_state_identifier->calc_String(*m_scope_Indexes[0], *m_scope_Indexes[1], bit);
	    break;
	}
	case 3:
	{
	    return m_state_identifier->calc_String(*m_scope_Indexes[0], *m_scope_Indexes[1], *m_scope_Indexes[2], bit);
	    break;
	}
	default:
	{
	    return m_state_identifier->calc_String(m_scope_Indexes, bit);
	    break;
	}
	}
    }


    int sSpecifiedStateIdentifier::calc_CNF(int bit) const
    {
	switch(m_scope_Indexes.size())
	{
	case 0:
	{
	    return m_state_identifier->calc_CNF(bit);
	    break;
	}
	case 1:
	{
	    return m_state_identifier->calc_CNF(*m_scope_Indexes[0], bit);
	    break;
	}
	case 2:
	{
	    return m_state_identifier->calc_CNF(*m_scope_Indexes[0], *m_scope_Indexes[1], bit);
	    break;
	}
	case 3:
	{
	    return m_state_identifier->calc_CNF(*m_scope_Indexes[0], *m_scope_Indexes[1], *m_scope_Indexes[2], bit);
	    break;
	}
	default:
	{
	    return m_state_identifier->calc_CNF(m_scope_Indexes, bit);
	    break;
	}
	}
    }
    

/*----------------------------------------------------------------------------*/
// sIndexableStateIdentifier

    sIndexableStateIdentifier::sIndexableStateIdentifier()
    {
	// nothing
    }
    
    sIndexableStateIdentifier::sIndexableStateIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, int N_States)
	: m_N_States(N_States)
	, m_indexable_identifier(variable_store, base_name, sIntegerScope(0, (m_log_N_States = calc_Log2(N_States)) - 1))
    {
	// nothing
    }


    sIndexableStateIdentifier::sIndexableStateIdentifier(sVariableStore_CNF *variable_store, const sString &base_name, int N_States, const sIndexScope &scope)
	: m_N_States(N_States)
	, m_indexable_identifier(variable_store, base_name, scope, sIntegerScope(0, (m_log_N_States = calc_Log2(N_States)) - 1))
    {
	// nothing
    }


    sIndexableStateIdentifier::sIndexableStateIdentifier(sVariableStore_CNF *variable_store,
							 const sString      &base_name,
							 int                 N_States,
							 const sIndexScope  &scope_1,
							 const sIndexScope  &scope_2)
	: m_N_States(N_States)
	, m_indexable_identifier(variable_store, base_name, scope_1, scope_2, sIntegerScope(0, (m_log_N_States = calc_Log2(N_States)) - 1))
    {
	// nothing
    }


    sIndexableStateIdentifier::sIndexableStateIdentifier(sVariableStore_CNF *variable_store,
							 const sString      &base_name,
							 int                 N_States,
							 const sIndexScope  &scope_1,
							 const sIndexScope  &scope_2,
							 const sIndexScope  &scope_3)
	: m_N_States(N_States)
    {
	IndexScopes_vector idx_Scopes;
	sIntegerScope state_scope(0, (m_log_N_States = calc_Log2(N_States)) - 1);
	idx_Scopes.push_back(&scope_1);
	idx_Scopes.push_back(&scope_2);
	idx_Scopes.push_back(&scope_3);
	idx_Scopes.push_back(&state_scope);

	m_indexable_identifier = sIndexableIdentifier(variable_store, base_name, idx_Scopes);
    }


    sIndexableStateIdentifier::sIndexableStateIdentifier(sVariableStore_CNF       *variable_store,
							 const sString            &base_name,
							 int                       N_States,
							 const IndexScopes_vector &index_Scopes)
	: m_N_States(N_States)
    {
	IndexScopes_vector idx_Scopes = index_Scopes;
	sIntegerScope state_scope(0, (m_log_N_States = calc_Log2(N_States)) - 1);
	idx_Scopes.push_back(&state_scope);

	m_indexable_identifier = sIndexableIdentifier(variable_store, base_name, idx_Scopes);
    }


    sIndexableStateIdentifier::sIndexableStateIdentifier(const sIndexableStateIdentifier &state_identifier)
	: m_N_States(state_identifier.m_N_States)
	, m_log_N_States(state_identifier.m_log_N_States)
	, m_indexable_identifier(state_identifier.m_indexable_identifier)
    {
	// nothing
    }


    const sIndexableStateIdentifier& sIndexableStateIdentifier::operator=(const sIndexableStateIdentifier &state_identifier)
    {
	m_N_States = state_identifier.m_N_States;
	m_log_N_States = state_identifier.m_log_N_States;
	m_indexable_identifier = state_identifier.m_indexable_identifier;

	return *this;
    }


    sIndexableStateIdentifier::~sIndexableStateIdentifier()
    {
	// nothing
    }


    sString sIndexableStateIdentifier::calc_String(int bit) const
    {
	return m_indexable_identifier.calc_String(sIntegerIndex(bit));
    }
    
    
    sString sIndexableStateIdentifier::calc_String(const sIndex &index, int bit) const
    {
	return m_indexable_identifier.calc_String(index, sIntegerIndex(bit));
    }
    
    
    sString sIndexableStateIdentifier::calc_String(const sIndex &index_1, const sIndex &index_2, int bit) const
    {
	return m_indexable_identifier.calc_String(index_1, index_2, sIntegerIndex(bit));
    }
    
    
    sString sIndexableStateIdentifier::calc_String(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3, int bit) const
    {
	Indexes_vector scope_Indexes;
	scope_Indexes.push_back(&index_1);
	scope_Indexes.push_back(&index_2);
	scope_Indexes.push_back(&index_3);

	sIntegerIndex bit_index(bit);
	scope_Indexes.push_back(&bit_index);

	return m_indexable_identifier.calc_String(scope_Indexes);
    }
    
    
    sString sIndexableStateIdentifier::calc_String(const Indexes_vector &scope_Indexes, int bit) const
    {
	Indexes_vector scope_Idxs = scope_Indexes;
	sIntegerIndex bit_index(bit);
	scope_Idxs.push_back(&bit_index);

	return m_indexable_identifier.calc_String(scope_Idxs);
    }
    

    int sIndexableStateIdentifier::calc_CNF(int bit) const
    {
	return m_indexable_identifier.calc_CNF(sIntegerIndex(bit));
    }
    
    
    int sIndexableStateIdentifier::calc_CNF(const sIndex &index, int bit) const
    {
	return m_indexable_identifier.calc_CNF(index, sIntegerIndex(bit));
    }
    
    
    int sIndexableStateIdentifier::calc_CNF(const sIndex &index_1, const sIndex &index_2, int bit) const
    {
	return m_indexable_identifier.calc_CNF(index_1, index_2, sIntegerIndex(bit));
    }
    
    
    int sIndexableStateIdentifier::calc_CNF(const sIndex &index_1, const sIndex &index_2, const sIndex &index_3, int bit) const
    {
	Indexes_vector scope_Indexes;
	scope_Indexes.push_back(&index_1);
	scope_Indexes.push_back(&index_2);
	scope_Indexes.push_back(&index_3);

	sIntegerIndex bit_index(bit);
	scope_Indexes.push_back(&bit_index);

	return m_indexable_identifier.calc_CNF(scope_Indexes);
    }


    int sIndexableStateIdentifier::calc_CNF(const Indexes_vector &scope_Indexes, int bit) const
    {
	Indexes_vector scope_Idxs = scope_Indexes;
	sIntegerIndex bit_index(bit);
	scope_Idxs.push_back(&bit_index);

	return m_indexable_identifier.calc_CNF(scope_Idxs);
    }


    sSpecifiedIdentifier sIndexableStateIdentifier::translate_CNFVariable(int cnf_variable) const
    {
	return m_indexable_identifier.translate_CNFVariable(cnf_variable);
    }


    void sIndexableStateIdentifier::to_Screen(const sString &indent) const
    {
	printf("%sIndexable identifier (|States| = %d, log_2|States| = %d) [\n", indent.c_str(), m_N_States, m_log_N_States);
	m_indexable_identifier.to_Screen(indent + sRELOC_INDENT);

	printf("%s]\n", indent.c_str());
    }


/*----------------------------------------------------------------------------*/
    
    int sIndexableStateIdentifier::calc_Log2(int value)
    {
	int log = 0;
	int v = 1;

	while (v < value)
	{
	    v *= 2;
	    ++log;
	}

	return log;
    }

    int sIndexableStateIdentifier::calc_Exp2(int value)
    {
	int exp = 1;

	while (value > 0)
	{
	    exp *= 2;
	    --value;
	}

	return exp;
    }


/*----------------------------------------------------------------------------*/
// sStateClauseGenerator

    sStateClauseGenerator::sStateClauseGenerator(sVariableStore_CNF *variable_store)
	: m_variable_store(variable_store)
	, m_aux_Identifier_cnt(0)
    {
	m_auxiliary_Identifiers[sINT_32_MAX] = sIndexableIdentifier();
    }


    sStateClauseGenerator::sStateClauseGenerator(const sStateClauseGenerator &clause_generator)
	: m_variable_store(clause_generator.m_variable_store)
	, m_aux_Identifier_cnt(clause_generator.m_aux_Identifier_cnt)
	, m_auxiliary_Identifiers(clause_generator.m_auxiliary_Identifiers)
    {
	// nothing
    }


    const sStateClauseGenerator& sStateClauseGenerator::operator=(const sStateClauseGenerator &clause_generator)
    {
	m_variable_store = clause_generator.m_variable_store;
	m_aux_Identifier_cnt = clause_generator.m_aux_Identifier_cnt;
	m_auxiliary_Identifiers = clause_generator.m_auxiliary_Identifiers;

	return *this;
    }


    int sStateClauseGenerator::count_Alignment(int &sUNUSED(aux_Variable_cnt), const sSpecifiedStateIdentifier &spec_identifier) const
    {
	int N_Bits = spec_identifier.get_StateIdentifier()->get_Log2_StateCount();
	int N_total_States = sIndexableStateIdentifier::calc_Exp2(N_Bits);
	int N_States = spec_identifier.get_StateIdentifier()->get_StateCount();

	return (N_total_States - N_States);
    }


    int sStateClauseGenerator::generate_Alignment(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, bool string)
    {
	int Clause_cnt = 0;
	int N_Bits = spec_identifier.get_StateIdentifier()->get_Log2_StateCount();

	int N_total_States = sIndexableStateIdentifier::calc_Exp2(N_Bits);
	int N_States = spec_identifier.get_StateIdentifier()->get_StateCount();

	for (int align_state = N_States; align_state < N_total_States; ++align_state)
	{
	    for (int bit = 0; bit < N_Bits; ++bit)
	    {
		// printf("align_state:%d\n", align_state);
		if (calc_Bit(bit, align_state) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier.calc_String(bit).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier.calc_CNF(bit));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier.calc_String(bit).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier.calc_CNF(bit));
		    }
		}
	    }
	    ++Clause_cnt;
	    fprintf(fw, " 0\n");
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_Equality(int &sUNUSED(aux_Variable_cnt), const sSpecifiedStateIdentifier &spec_identifier, int sUNUSED(state)) const
    {
	return spec_identifier.get_StateIdentifier()->get_Log2_StateCount();
    }


    int sStateClauseGenerator::generate_Equality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, int state, bool string)
    {
	int Clause_cnt = 0;
	int N_Bits = spec_identifier.get_StateIdentifier()->get_Log2_StateCount();

	for (int bit = 0; bit < N_Bits; ++bit)
	{
	    int state_bit = calc_Bit(bit, state);
	    sASSERT(state_bit == 1 || state_bit == 0);

	    if (state_bit == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier.calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier.calc_CNF(bit));
		}
		++Clause_cnt;
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier.calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier.calc_CNF(bit));
		}
		++Clause_cnt;
	    }
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_Disequality(int &sUNUSED(aux_Variable_cnt), const sSpecifiedStateIdentifier &sUNUSED(spec_identifier), int sUNUSED(state)) const
    {
	return 1;
    }


    int sStateClauseGenerator::generate_Disequality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier, int state, bool string)
    {
	int Clause_cnt = 0;
	int N_Bits = spec_identifier.get_StateIdentifier()->get_Log2_StateCount();

	for (int bit = 0; bit < N_Bits; ++bit)
	{
	    int state_bit = calc_Bit(bit, state);
	    sASSERT(state_bit == 1 || state_bit == 0);

	    if (state_bit == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier.calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier.calc_CNF(bit));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier.calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier.calc_CNF(bit));
		}
	    }
	}
	fprintf(fw, "0\n");
	++Clause_cnt;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_Equality(int &sUNUSED(aux_Variable_cnt), const sSpecifiedStateIdentifier &spec_identifier_A, const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_B)) const
    {
	return 2 * spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();
    }


    int sStateClauseGenerator::generate_Equality(FILE *fw, const sSpecifiedStateIdentifier &spec_identifier_A, const sSpecifiedStateIdentifier &spec_identifier_B, bool string)
    {
	sASSERT(spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_B.get_StateIdentifier()->get_Log2_StateCount());

	int Clause_cnt = 0;
	int N_Bits = spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();

	for (int bit = 0; bit < N_Bits; ++bit)
	{	   
	    if (string)
	    {
		fprintf(fw, "-%s %s 0\n", spec_identifier_A.calc_String(bit).c_str(), spec_identifier_B.calc_String(bit).c_str());
		fprintf(fw, "%s -%s 0\n", spec_identifier_A.calc_String(bit).c_str(), spec_identifier_B.calc_String(bit).c_str());
	    }
	    else
	    {
		fprintf(fw, "-%d %d 0\n", spec_identifier_A.calc_CNF(bit), spec_identifier_B.calc_CNF(bit));
		fprintf(fw, "%d -%d 0\n", spec_identifier_A.calc_CNF(bit), spec_identifier_B.calc_CNF(bit));
	    }
	    Clause_cnt += 2;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B)) const
    {
	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_IF_Bits + N_THEN_Bits;

	return (4 * N_IF_Bits + 5 * N_THEN_Bits);
    }


    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
							    bool                             string)
    {
	sASSERT(spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_IF_B.get_StateIdentifier()->get_Log2_StateCount())
	sASSERT(spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier not_equal_auxiliary(m_variable_store, "neq_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_IF_Bits - 1));
	m_auxiliary_Identifiers[not_equal_auxiliary.get_First_CNFVariable()] = not_equal_auxiliary;

	sIndexableIdentifier equal_auxiliary(m_variable_store, "eql_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_Bits - 1));
	m_auxiliary_Identifiers[equal_auxiliary.get_First_CNFVariable()] = equal_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (string)
		{
		    fprintf(fw, "%s ", not_equal_auxiliary.calc_String(sIntegerIndex(bit_2)).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", not_equal_auxiliary.calc_CNF(sIntegerIndex(bit_2)));
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_IF_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s -%s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s -%s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s %s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s %s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d -%d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d -%d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d %d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d %d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}

	for (int bit = 0; bit < N_THEN_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &sUNUSED(aux_Variable_cnt),
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							 int                              sUNUSED(state_THEN_B)) const
    {
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	return N_THEN_Bits;
    }

	
    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							    int                              state_THEN_B,
							    bool                             string)
    {
	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	for (int bit_1 = 0; bit_1 < N_THEN_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B)) const
    {
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_THEN_Bits;

	return (5 * N_THEN_Bits);
    }
	

    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
							    bool                             string)
    {
	sASSERT(spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier equal_auxiliary(m_variable_store, "eql_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_Bits - 1));
	m_auxiliary_Identifiers[equal_auxiliary.get_First_CNFVariable()] = equal_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_THEN_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A.calc_String(bit).c_str(), spec_identifier_THEN_B.calc_String(bit).c_str(), equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A.calc_CNF(bit), spec_identifier_THEN_B.calc_CNF(bit), equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_1),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_2)) const
    {
	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_IF_Bits + N_THEN_1_Bits + N_THEN_2_Bits;

	return (4 * N_IF_Bits + 5 * N_THEN_1_Bits + 5 * N_THEN_2_Bits);
    }


    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
							    bool                             string)
    {
	sASSERT(spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_IF_B.get_StateIdentifier()->get_Log2_StateCount())
	sASSERT(spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_1.get_StateIdentifier()->get_Log2_StateCount())
	sASSERT(spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_2.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier not_equal_auxiliary(m_variable_store, "neq_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_IF_Bits - 1));
	m_auxiliary_Identifiers[not_equal_auxiliary.get_First_CNFVariable()] = not_equal_auxiliary;

	sIndexableIdentifier equal_1_auxiliary(m_variable_store, "eql_1_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_1_Bits - 1));
	m_auxiliary_Identifiers[equal_1_auxiliary.get_First_CNFVariable()] = equal_1_auxiliary;

	sIndexableIdentifier equal_2_auxiliary(m_variable_store, "eql_2_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_2_Bits - 1));
	m_auxiliary_Identifiers[equal_2_auxiliary.get_First_CNFVariable()] = equal_2_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (string)
		{
		    fprintf(fw, "%s ", not_equal_auxiliary.calc_String(sIntegerIndex(bit_2)).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", not_equal_auxiliary.calc_CNF(sIntegerIndex(bit_2)));
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_1_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_1_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (string)
		{
		    fprintf(fw, "%s ", not_equal_auxiliary.calc_String(sIntegerIndex(bit_2)).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", not_equal_auxiliary.calc_CNF(sIntegerIndex(bit_2)));
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_2_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_2_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_IF_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s -%s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s -%s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s %s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s %s 0\n", spec_identifier_IF_A.calc_String(bit).c_str(), spec_identifier_IF_B.calc_String(bit).c_str(), not_equal_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d -%d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d -%d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d %d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d %d 0\n", spec_identifier_IF_A.calc_CNF(bit), spec_identifier_IF_B.calc_CNF(bit), not_equal_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}

	for (int bit = 0; bit < N_THEN_1_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A_1.calc_String(bit).c_str(), spec_identifier_THEN_B_1.calc_String(bit).c_str(), equal_1_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A_1.calc_String(bit).c_str(), spec_identifier_THEN_B_1.calc_String(bit).c_str(), equal_1_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A_1.calc_String(bit).c_str(), spec_identifier_THEN_B_1.calc_String(bit).c_str(), equal_1_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A_1.calc_String(bit).c_str(), spec_identifier_THEN_B_1.calc_String(bit).c_str(), equal_1_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit), spec_identifier_THEN_B_1.calc_CNF(bit), equal_1_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit), spec_identifier_THEN_B_1.calc_CNF(bit), equal_1_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit), spec_identifier_THEN_B_1.calc_CNF(bit), equal_1_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit), spec_identifier_THEN_B_1.calc_CNF(bit), equal_1_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}

	for (int bit = 0; bit < N_THEN_2_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &sUNUSED(aux_Variable_cnt),
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							 int                              sUNUSED(state_THEN_B_1),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							 int                              sUNUSED(state_THEN_B_2)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	return (N_THEN_1_Bits + N_THEN_2_Bits);
    }
	

    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              state_THEN_B_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    int                              state_THEN_B_2,
							    bool                             string)
    {
	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_2) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							 int                              sUNUSED(state_THEN_B_1),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_2)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_THEN_2_Bits;

	return (N_THEN_1_Bits + 5 * N_THEN_2_Bits);
    }
	

    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              state_THEN_B_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
							    bool                             string)
    {
	sASSERT(spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_2.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier equal_2_auxiliary(m_variable_store, "eql_2_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_2_Bits - 1));
	m_auxiliary_Identifiers[equal_2_auxiliary.get_First_CNFVariable()] = equal_2_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_2_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_2_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_THEN_2_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A_2.calc_String(bit).c_str(), spec_identifier_THEN_B_2.calc_String(bit).c_str(), equal_2_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit), spec_identifier_THEN_B_2.calc_CNF(bit), equal_2_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							 int                              sUNUSED(state_THEN_B_1),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							 int                              sUNUSED(state_THEN_B_2),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_3)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_THEN_3_Bits;

	return (N_THEN_1_Bits + N_THEN_2_Bits + 5 * N_THEN_3_Bits);
    }
	

    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              state_THEN_B_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    int                              state_THEN_B_2,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B_3,
							    bool                             string)
    {
	sASSERT(spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_3.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier equal_3_auxiliary(m_variable_store, "eql_3_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_3_Bits - 1));
	m_auxiliary_Identifiers[equal_3_auxiliary.get_First_CNFVariable()] = equal_3_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_2) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_3_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_3_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_3_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_THEN_3_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A_3.calc_String(bit).c_str(), spec_identifier_THEN_B_3.calc_String(bit).c_str(), equal_3_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A_3.calc_String(bit).c_str(), spec_identifier_THEN_B_3.calc_String(bit).c_str(), equal_3_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A_3.calc_String(bit).c_str(), spec_identifier_THEN_B_3.calc_String(bit).c_str(), equal_3_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A_3.calc_String(bit).c_str(), spec_identifier_THEN_B_3.calc_String(bit).c_str(), equal_3_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit), spec_identifier_THEN_B_3.calc_CNF(bit), equal_3_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit), spec_identifier_THEN_B_3.calc_CNF(bit), equal_3_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit), spec_identifier_THEN_B_3.calc_CNF(bit), equal_3_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit), spec_identifier_THEN_B_3.calc_CNF(bit), equal_3_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							 int                              sUNUSED(state_IF_B),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							 int                              sUNUSED(state_THEN_B_1),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							 int                              sUNUSED(state_THEN_B_2),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							 int                              sUNUSED(state_THEN_B_3),
							 const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
							 const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_4)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_4_Bits = spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_THEN_4_Bits;

	return (N_THEN_1_Bits + N_THEN_2_Bits + N_THEN_3_Bits + 5 * N_THEN_4_Bits);
    }
	

    int sStateClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							    const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							    int                              state_IF_B,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              state_THEN_B_1,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    int                              state_THEN_B_2,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							    int                              state_THEN_B_3,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4,
							    bool                             string)
    {
	sASSERT(spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_4.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_4_Bits = spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier equal_4_auxiliary(m_variable_store, "eql_4_aux-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_THEN_4_Bits - 1));
	m_auxiliary_Identifiers[equal_4_auxiliary.get_First_CNFVariable()] = equal_4_auxiliary;

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_2) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_3_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (calc_Bit(bit_1, state_THEN_B_3) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "%s 0\n", spec_identifier_THEN_A_3.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s 0\n", spec_identifier_THEN_A_3.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d 0\n", spec_identifier_THEN_A_3.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_4_Bits; ++bit_1)
	{
	    for (int bit_2 = 0; bit_2 < N_IF_Bits; ++bit_2)
	    {
		if (calc_Bit(bit_2, state_IF_B) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_2).c_str());
		    }
		    else
		    {
			fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_2));
		    }
		}
	    }
	    if (string)
	    {
		fprintf(fw, "%s 0\n", equal_4_auxiliary.calc_String(sIntegerIndex(bit_1)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d 0\n", equal_4_auxiliary.calc_CNF(sIntegerIndex(bit_1)));
	    }
	    ++Clause_cnt;
	}

	for (int bit = 0; bit < N_THEN_4_Bits; ++bit)
	{
	    if (string)
	    {
		fprintf(fw, "%s %s %s 0\n", spec_identifier_THEN_A_4.calc_String(bit).c_str(), spec_identifier_THEN_B_4.calc_String(bit).c_str(), equal_4_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s -%s %s 0\n", spec_identifier_THEN_A_4.calc_String(bit).c_str(), spec_identifier_THEN_B_4.calc_String(bit).c_str(), equal_4_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "-%s %s -%s 0\n", spec_identifier_THEN_A_4.calc_String(bit).c_str(), spec_identifier_THEN_B_4.calc_String(bit).c_str(), equal_4_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		fprintf(fw, "%s -%s -%s 0\n", spec_identifier_THEN_A_4.calc_String(bit).c_str(), spec_identifier_THEN_B_4.calc_String(bit).c_str(), equal_4_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d %d %d 0\n", spec_identifier_THEN_A_4.calc_CNF(bit), spec_identifier_THEN_B_4.calc_CNF(bit), equal_4_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d -%d %d 0\n", spec_identifier_THEN_A_4.calc_CNF(bit), spec_identifier_THEN_B_4.calc_CNF(bit), equal_4_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "-%d %d -%d 0\n", spec_identifier_THEN_A_4.calc_CNF(bit), spec_identifier_THEN_B_4.calc_CNF(bit), equal_4_auxiliary.calc_CNF(sIntegerIndex(bit)));
		fprintf(fw, "%d -%d -%d 0\n", spec_identifier_THEN_A_4.calc_CNF(bit), spec_identifier_THEN_B_4.calc_CNF(bit), equal_4_auxiliary.calc_CNF(sIntegerIndex(bit)));
	    }
	    Clause_cnt += 4;
	}
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_DifferenceConstraint(int                              &aux_Variable_cnt,
							  const sSpecifiedStateIdentifier  &spec_identifier_A,
							  SpecifiedStateIdentifiers_vector &spec_Identifiers_B) const
    {
	int N_Bits = spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_Identifiers = spec_Identifiers_B.size();
	aux_Variable_cnt += N_Identifiers * N_Bits;

	return N_Identifiers * (1 + 2 * N_Bits);
    }
	

    int sStateClauseGenerator::generate_DifferenceConstraint(FILE                             *fw,
							     const sSpecifiedStateIdentifier  &spec_identifier_A,
							     SpecifiedStateIdentifiers_vector &spec_Identifiers_B,
							     bool                              string)
    {
	#ifdef sDEBUG
	{
	    for (SpecifiedStateIdentifiers_vector::const_iterator identifier_B = spec_Identifiers_B.begin(); identifier_B != spec_Identifiers_B.end(); ++identifier_B)
	    {
		sASSERT(spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount() == identifier_B->get_StateIdentifier()->get_Log2_StateCount());
	    }
	}
	#endif

	int Clause_cnt = 0;
	int N_Bits = spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();

	for (SpecifiedStateIdentifiers_vector::const_iterator identifier_B = spec_Identifiers_B.begin(); identifier_B != spec_Identifiers_B.end(); ++identifier_B)
	{
	    sIndexableIdentifier diff_auxiliary(m_variable_store, "diff_aux_" +  sInt_32_to_String(identifier_B - spec_Identifiers_B.begin()) + "-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_Bits - 1));
	    m_auxiliary_Identifiers[diff_auxiliary.get_First_CNFVariable()] = diff_auxiliary;

	    for (int bit = 0; bit < N_Bits; ++bit)
	    {
		if (string)
		{
		    fprintf(fw, "%s ", diff_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", diff_auxiliary.calc_CNF(sIntegerIndex(bit)));
		}
	    }
	    fprintf(fw, "0\n");
	    ++Clause_cnt;

	    for (int bit = 0; bit < N_Bits; ++bit)
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s -%s 0\n", spec_identifier_A.calc_String(bit).c_str(), identifier_B->calc_String(bit).c_str(), diff_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		    fprintf(fw, "%s %s -%s 0\n", spec_identifier_A.calc_String(bit).c_str(), identifier_B->calc_String(bit).c_str(), diff_auxiliary.calc_String(sIntegerIndex(bit)).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d -%d 0\n", spec_identifier_A.calc_CNF(bit), identifier_B->calc_CNF(bit), diff_auxiliary.calc_CNF(sIntegerIndex(bit)));
		    fprintf(fw, "%d %d -%d 0\n", spec_identifier_A.calc_CNF(bit), identifier_B->calc_CNF(bit), diff_auxiliary.calc_CNF(sIntegerIndex(bit)));
		}
	    }
	    Clause_cnt += 2;
	}

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_AllDifferenceConstraint(int                              &aux_Variable_cnt,
							     SpecifiedStateIdentifiers_vector &spec_Identifiers) const
    {
	int N_Bits = spec_Identifiers[0].get_StateIdentifier()->get_Log2_StateCount();
	int N_Identifiers = spec_Identifiers.size();
	aux_Variable_cnt += N_Identifiers * (N_Identifiers - 1) * N_Bits;

	return N_Identifiers * (N_Identifiers - 1) * (1 + 2 * N_Bits);
    }


    int sStateClauseGenerator::generate_AllDifferenceConstraint(FILE                             *fw,
								SpecifiedStateIdentifiers_vector &spec_Identifiers,
								bool                              string)
    {
	int Clause_cnt = 0;

	for (SpecifiedStateIdentifiers_vector::const_iterator identifier = spec_Identifiers.begin(); identifier != spec_Identifiers.end(); ++identifier)
	{
	    SpecifiedStateIdentifiers_vector diff_spec_Identifiers;
	 
	    SpecifiedStateIdentifiers_vector::const_iterator diff_identifier;
	    for (diff_identifier = spec_Identifiers.begin(); diff_identifier != identifier; ++diff_identifier)
	    {
		diff_spec_Identifiers.push_back(*diff_identifier);
	    }
	    for (++diff_identifier; diff_identifier != spec_Identifiers.end(); ++diff_identifier)
	    {
		diff_spec_Identifiers.push_back(*diff_identifier);
	    }	   
	    Clause_cnt += generate_DifferenceConstraint(fw, *identifier, diff_spec_Identifiers, string);
	}

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_CaseSplitting(int                              &aux_Variable_cnt,
						   SpecifiedStateIdentifiers_vector &split_Identifiers,
						   States_vector                    &sUNUSED(split_States)) const
    {
	int N_Splits = split_Identifiers.size();
	SpecifiedStateIdentifiers_vector::const_iterator split_identifier = split_Identifiers.begin();
	int N_Bits = split_identifier->get_StateIdentifier()->get_Log2_StateCount();

	#ifdef sDEBUG
	{
	    SpecifiedStateIdentifiers_vector::const_iterator split_identifier = split_Identifiers.begin();
	    int N_Bits = split_identifier->get_StateIdentifier()->get_Log2_StateCount();

	    for (++split_identifier; split_identifier != split_Identifiers.end(); ++split_identifier)
	    {
		sASSERT(N_Bits == split_identifier->get_StateIdentifier()->get_Log2_StateCount());
	    }
	}
	#endif

	aux_Variable_cnt += N_Splits;

	return 2 + (N_Splits - 1) * N_Bits;
    }


    int sStateClauseGenerator::generate_CaseSplitting(FILE                             *fw,
						      SpecifiedStateIdentifiers_vector &split_Identifiers,
						      States_vector                    &split_States,
						      bool                              string)
    {
	int Clause_cnt = 0;
	int N_Splits = split_Identifiers.size();
	
	sIndexableIdentifier split_auxiliary(m_variable_store, "split_aux_" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_Splits - 1));
	m_auxiliary_Identifiers[split_auxiliary.get_First_CNFVariable()] = split_auxiliary;

	SpecifiedStateIdentifiers_vector::const_iterator split_identifier = split_Identifiers.begin();
	int N_Bits = split_identifier->get_StateIdentifier()->get_Log2_StateCount();

	if (string)
	{
	    fprintf(fw, "-%s ", split_auxiliary.calc_String(sIntegerIndex(0)).c_str());
	}
	else
	{
	    fprintf(fw, "-%d ", split_auxiliary.calc_CNF(sIntegerIndex(0)));
	}
	
	for (int bit = 0; bit < N_Bits; ++bit)
	{
	    if (calc_Bit(bit, split_States[0]) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", split_identifier->calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", split_identifier->calc_CNF(bit));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", split_identifier->calc_String(bit).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", split_identifier->calc_CNF(bit));
		}
	    }
	}

	fprintf(fw, "0\n");
	++Clause_cnt;

	for (++split_identifier; split_identifier != split_Identifiers.end(); ++split_identifier)
	{
	    int N_Bits = split_identifier->get_StateIdentifier()->get_Log2_StateCount();
	    	    
	    for (int bit = 0; bit < N_Bits; ++bit)
	    {
		if (calc_Bit(bit, split_States[split_identifier - split_Identifiers.begin()]) == 1)
		{
		    if (string)
		    {
			fprintf(fw, "-%s %s 0\n", split_auxiliary.calc_String(sIntegerIndex(split_identifier - split_Identifiers.begin())).c_str(), split_identifier->calc_String(bit).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d %d 0\n", split_auxiliary.calc_CNF(sIntegerIndex(split_identifier - split_Identifiers.begin())), split_identifier->calc_CNF(bit));
		    }
		    ++Clause_cnt;
		}
		else
		{
		    if (string)
		    {
			fprintf(fw, "-%s -%s 0\n", split_auxiliary.calc_String(sIntegerIndex(split_identifier - split_Identifiers.begin())).c_str(), split_identifier->calc_String(bit).c_str());
		    }
		    else
		    {
			fprintf(fw, "-%d -%d 0\n", split_auxiliary.calc_CNF(sIntegerIndex(split_identifier - split_Identifiers.begin())), split_identifier->calc_CNF(bit));
		    }
		    ++Clause_cnt;
		}
	    }
	}

	for (int split = 0; split < N_Splits; ++split)
	{
	    if (string)
	    {
		fprintf(fw, "%s ", split_auxiliary.calc_String(sIntegerIndex(split)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d ", split_auxiliary.calc_CNF(sIntegerIndex(split)));
	    }
	}

	fprintf(fw, "0\n");
	++Clause_cnt;

	return Clause_cnt;
    }


    int sStateClauseGenerator::count_LargeConditionalEquality(int                              &sUNUSED(aux_Variable_cnt),
							      const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
							      int                               sUNUSED(state_IF_B),
							      SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_A),
							      SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B))
    {
	sASSERT(false);
	return 0;
    }

    
    int sStateClauseGenerator::generate_LargeConditionalEquality(FILE                             *sUNUSED(fw),
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
								 int                               sUNUSED(state_IF_B),
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_A),
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B),
								 bool                              sUNUSED(string))
    {
	sASSERT(false);
	return 0;
    }
    

    int sStateClauseGenerator::count_LargeConditionalEquality(int                              &sUNUSED(aux_Variable_cnt),
							      const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
							      int                               sUNUSED(state_IF_B),
							      const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_THEN_A_1),
							      int                               sUNUSED(state_THEN_B_1),
							      const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_THEN_A_2),
							      int                               sUNUSED(state_THEN_B_2),
							      SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_A_3),
							      SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B_3))
    {
	sASSERT(false);
	return 0;
    }
    

    int sStateClauseGenerator::generate_LargeConditionalEquality(FILE                             *sUNUSED(fw),
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
								 int                               sUNUSED(state_IF_B),
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_THEN_A_1),
								 int                               sUNUSED(state_THEN_B_1),
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_THEN_A_2),
								 int                               sUNUSED(state_THEN_B_2),
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_A_3),
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B_3),
								 bool                              sUNUSED(string))
    {
	sASSERT(false);
	return 0;
    }


    sSpecifiedIdentifier sStateClauseGenerator::translate_AuxiliaryCNFVariable(int cnf_variable) const
    {
	AuxiliaryIdentifiers_map::const_iterator auxiliary_identifier = m_auxiliary_Identifiers.upper_bound(cnf_variable);

	/*
	for (AuxiliaryIdentifiers_map::const_iterator identifier = m_auxiliary_Identifiers.begin(); identifier != m_auxiliary_Identifiers.end(); ++identifier)
	{
	    printf("First var:%d\n", identifier->first);
	    identifier->second.to_Screen();
	}
	//	exit(0);
	*/
       
	if (auxiliary_identifier != m_auxiliary_Identifiers.end())
	{
	    if (auxiliary_identifier->first > cnf_variable)
	    {
		--auxiliary_identifier;
	    }
	    if (auxiliary_identifier->second.is_Anonymous())
	    {
		return sSpecifiedIdentifier();
	    }
	    else
	    {
		return auxiliary_identifier->second.translate_CNFVariable(cnf_variable);
	    }
	}

	return sSpecifiedIdentifier();
    }

    int sStateClauseGenerator::count_LEXLess_Constraint(int                             &aux_Variable_cnt,
							const sSpecifiedStateIdentifier &spec_identifier_A,
							const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_B)) const
    {
	int N_Bits = spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += N_Bits;

	return (1 + (N_Bits * (N_Bits + 1)));
    }

	
    int sStateClauseGenerator::generate_LEXLess_Constraint(FILE                            *fw,
							   const sSpecifiedStateIdentifier &spec_identifier_A,
							   const sSpecifiedStateIdentifier &spec_identifier_B,
							   bool                             string)
    {
	int Clause_cnt = 0;

	sASSERT(spec_identifier_A.get_StateIdentifier()->get_StateCount() == spec_identifier_B.get_StateIdentifier()->get_StateCount());

	int N_Bits = spec_identifier_A.get_StateIdentifier()->get_Log2_StateCount();

	sIndexableIdentifier lex_less_auxiliary(m_variable_store, "lex_less-" + sInt_32_to_String(m_aux_Identifier_cnt++), sIntegerScope(0, N_Bits - 1));
	m_auxiliary_Identifiers[lex_less_auxiliary.get_First_CNFVariable()] = lex_less_auxiliary;

	for (int diff_pos = N_Bits - 1; diff_pos >= 0; --diff_pos)
	{
	    for (int equal_bit = N_Bits - 1; equal_bit > diff_pos; --equal_bit)
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s %s 0\n", lex_less_auxiliary.calc_String(sIntegerIndex(diff_pos)).c_str(), spec_identifier_A.calc_String(equal_bit).c_str(), spec_identifier_B.calc_String(equal_bit).c_str());
		    fprintf(fw, "-%s %s -%s 0\n", lex_less_auxiliary.calc_String(sIntegerIndex(diff_pos)).c_str(), spec_identifier_A.calc_String(equal_bit).c_str(), spec_identifier_B.calc_String(equal_bit).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d %d 0\n", lex_less_auxiliary.calc_CNF(sIntegerIndex(diff_pos)), spec_identifier_A.calc_CNF(equal_bit), spec_identifier_B.calc_CNF(equal_bit));
		    fprintf(fw, "-%d %d -%d 0\n", lex_less_auxiliary.calc_CNF(sIntegerIndex(diff_pos)), spec_identifier_A.calc_CNF(equal_bit), spec_identifier_B.calc_CNF(equal_bit));
		}
		Clause_cnt += 2;
	    }
	    if (string)
	    {
		fprintf(fw, "-%s -%s 0\n", lex_less_auxiliary.calc_String(sIntegerIndex(diff_pos)).c_str(), spec_identifier_A.calc_String(diff_pos).c_str());
		fprintf(fw, "-%s %s 0\n", lex_less_auxiliary.calc_String(sIntegerIndex(diff_pos)).c_str(), spec_identifier_B.calc_String(diff_pos).c_str());
	    }
	    else
	    {
		fprintf(fw, "-%d -%d 0\n", lex_less_auxiliary.calc_CNF(sIntegerIndex(diff_pos)), spec_identifier_A.calc_CNF(diff_pos));
		fprintf(fw, "-%d %d 0\n", lex_less_auxiliary.calc_CNF(sIntegerIndex(diff_pos)), spec_identifier_B.calc_CNF(diff_pos));
	    }
	    Clause_cnt += 2;
	}

	for (int diff_pos = N_Bits - 1; diff_pos >= 0; --diff_pos)
	{
	    if (string)
	    {
		fprintf(fw, "%s ", lex_less_auxiliary.calc_String(sIntegerIndex(diff_pos)).c_str());
	    }
	    else
	    {
		fprintf(fw, "%d ", lex_less_auxiliary.calc_CNF(sIntegerIndex(diff_pos)));
	    }
	}
	fprintf(fw, "0\n");

	++Clause_cnt;

	return Clause_cnt;
    }


/*----------------------------------------------------------------------------*/

    int sStateClauseGenerator::calc_Bit(int bit, int state)
    {
	while (bit-- > 0)
	{
	    state /= 2;
	}
	return state % 2;
    }


/*----------------------------------------------------------------------------*/
// sAdvancedClauseGenerator

    sAdvancedClauseGenerator::sAdvancedClauseGenerator(sVariableStore_CNF *variable_store)
	: sStateClauseGenerator(variable_store)
    {
	// nothing
    }


    sAdvancedClauseGenerator::sAdvancedClauseGenerator(const sAdvancedClauseGenerator &clause_generator)
	: sStateClauseGenerator(clause_generator)
    {
	// nothing
    }


    const sAdvancedClauseGenerator& sAdvancedClauseGenerator::operator=(const sAdvancedClauseGenerator &clause_generator)
    {
	sStateClauseGenerator::operator=(clause_generator);
	return *this;
    }


    int sAdvancedClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							    int                              sUNUSED(state_IF_B),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B)) const
    {
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += 1;

	return (1 + 2 * N_THEN_Bits);
    }
	

    int sAdvancedClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							       const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							       int                              state_IF_B,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_B,
							       bool                             string)
    {
	sASSERT(spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;

	for (int bit_2 = 0; bit_2 < N_THEN_Bits; ++bit_2)
	{
	    if (string)
	    {
		fprintf(fw, "-%s -%s %s 0\n", "auxiliary", spec_identifier_THEN_A.calc_String(bit_2).c_str(), spec_identifier_THEN_B.calc_String(bit_2).c_str());
		fprintf(fw, "-%s %s -%s 0\n", "auxiliary", spec_identifier_THEN_A.calc_String(bit_2).c_str(), spec_identifier_THEN_B.calc_String(bit_2).c_str());
	    }
	    else
	    {
		fprintf(fw, "-%d -%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A.calc_CNF(bit_2), spec_identifier_THEN_B.calc_CNF(bit_2));
		fprintf(fw, "-%d %d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A.calc_CNF(bit_2), spec_identifier_THEN_B.calc_CNF(bit_2));
	    }
	    Clause_cnt += 2;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }

    int sAdvancedClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							    int                              sUNUSED(state_IF_B),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							    int                              sUNUSED(state_THEN_B)) const
    {
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += 1;

	return (1 + N_THEN_Bits);
    }
	

    int sAdvancedClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							       const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							       int                              state_IF_B,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A,
							       int                              state_THEN_B,
							       bool                             string)
    {
	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_Bits = spec_identifier_THEN_A.get_StateIdentifier()->get_Log2_StateCount();

	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;


	for (int bit_1 = 0; bit_1 < N_THEN_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							    int                              sUNUSED(state_IF_B),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              sUNUSED(state_THEN_B_1),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_2)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += 1;

	return (1 + N_THEN_1_Bits + 2 * N_THEN_2_Bits);
    }
	

    int sAdvancedClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							       const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							       int                              state_IF_B,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							       int                              state_THEN_B_1,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_B_2,
							       bool                             string)
    {
	sASSERT(spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_2.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;

	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    if (string)
	    {
		fprintf(fw, "-%s -%s %s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str(), spec_identifier_THEN_B_2.calc_String(bit_1).c_str());
		fprintf(fw, "-%s %s -%s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str(), spec_identifier_THEN_B_2.calc_String(bit_1).c_str());
	    }
	    else
	    {
		fprintf(fw, "-%d -%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1), spec_identifier_THEN_B_2.calc_CNF(bit_1));
		fprintf(fw, "-%d %d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1), spec_identifier_THEN_B_2.calc_CNF(bit_1));
	    }
	    Clause_cnt += 2;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::count_ConditionalEquality(int                             &aux_Variable_cnt,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_IF_A),
							    int                              sUNUSED(state_IF_B),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							    int                              sUNUSED(state_THEN_B_1),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							    int                              sUNUSED(state_THEN_B_2),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							    int                              sUNUSED(state_THEN_B_3),
							    const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
							    const sSpecifiedStateIdentifier &sUNUSED(spec_identifier_THEN_B_4)) const
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_4_Bits = spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount();

	aux_Variable_cnt += 1;

	return (1 + N_THEN_1_Bits + N_THEN_2_Bits + N_THEN_3_Bits + 2 * N_THEN_4_Bits);
    }
	

    int sAdvancedClauseGenerator::generate_ConditionalEquality(FILE                            *fw,
							       const sSpecifiedStateIdentifier &spec_identifier_IF_A,
							       int                              state_IF_B,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_1,
							       int                              state_THEN_B_1,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_2,
							       int                              state_THEN_B_2,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_3,
							       int                              state_THEN_B_3,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_A_4,
							       const sSpecifiedStateIdentifier &spec_identifier_THEN_B_4,
							       bool                             string)
    {
	sASSERT(spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount() == spec_identifier_THEN_B_4.get_StateIdentifier()->get_Log2_StateCount())

	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_3_Bits = spec_identifier_THEN_A_3.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_4_Bits = spec_identifier_THEN_A_4.get_StateIdentifier()->get_Log2_StateCount();

	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;


	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_2) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_3_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_3) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_3.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_3.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_3.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_3.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_4_Bits; ++bit_1)
	{
	    if (string)
	    {
		fprintf(fw, "-%s -%s %s 0\n", "auxiliary", spec_identifier_THEN_A_4.calc_String(bit_1).c_str(), spec_identifier_THEN_B_4.calc_String(bit_1).c_str());
		fprintf(fw, "-%s %s -%s 0\n", "auxiliary", spec_identifier_THEN_A_4.calc_String(bit_1).c_str(), spec_identifier_THEN_B_4.calc_String(bit_1).c_str());
	    }
	    else
	    {
		fprintf(fw, "-%d -%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_4.calc_CNF(bit_1), spec_identifier_THEN_B_4.calc_CNF(bit_1));
		fprintf(fw, "-%d %d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_4.calc_CNF(bit_1), spec_identifier_THEN_B_4.calc_CNF(bit_1));
	    }
	    Clause_cnt += 2;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
	#endif

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::count_LargeConditionalEquality(int                              &aux_Variable_cnt,
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
								 int                               sUNUSED(state_IF_B),
								 SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B))
    {
	int N_THEN_Bits = 0;
	int N_Equalities = spec_Identifiers_THEN_A.size();
	
	for (int equality = 0; equality < N_Equalities; ++equality)
	{
	    N_THEN_Bits += spec_Identifiers_THEN_A[equality].get_StateIdentifier()->get_Log2_StateCount();
	}
	aux_Variable_cnt += 1;

	return (1 + 2 * N_THEN_Bits);
    }


    int sAdvancedClauseGenerator::generate_LargeConditionalEquality(FILE                             *fw,
								    const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
								    int                               state_IF_B,
								    SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A,
								    SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B,
								    bool                              string)
    {
	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;

	int N_Equalities = spec_Identifiers_THEN_A.size();
	sASSERT(spec_Identifiers_THEN_A.size() == spec_Identifiers_THEN_B.size());
	
	for (int equality = 0; equality < N_Equalities; ++equality)
	{
	    sASSERT(spec_Identifiers_THEN_A[equality].get_StateIdentifier()->get_Log2_StateCount() == spec_Identifiers_THEN_B[equality].get_StateIdentifier()->get_Log2_StateCount());
	    int N_THEN_Bits = spec_Identifiers_THEN_A[equality].get_StateIdentifier()->get_Log2_StateCount();
	    
	    for (int bit_1 = 0; bit_1 < N_THEN_Bits; ++bit_1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s %s 0\n", "auxiliary", spec_Identifiers_THEN_A[equality].calc_String(bit_1).c_str(), spec_Identifiers_THEN_B[equality].calc_String(bit_1).c_str());
		    fprintf(fw, "-%s %s -%s 0\n", "auxiliary", spec_Identifiers_THEN_A[equality].calc_String(bit_1).c_str(), spec_Identifiers_THEN_B[equality].calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d %d 0\n", aux_cnf_variable, spec_Identifiers_THEN_A[equality].calc_CNF(bit_1), spec_Identifiers_THEN_B[equality].calc_CNF(bit_1));
		    fprintf(fw, "-%d %d -%d 0\n", aux_cnf_variable, spec_Identifiers_THEN_A[equality].calc_CNF(bit_1), spec_Identifiers_THEN_B[equality].calc_CNF(bit_1));
		}
		Clause_cnt += 2;
	    }
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
        #endif

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::count_LargeConditionalEquality(int                              &aux_Variable_cnt,
								 const sSpecifiedStateIdentifier  &sUNUSED(spec_identifier_IF_A),
								 int                               sUNUSED(state_IF_B),
								 const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
								 int                               sUNUSED(state_THEN_B_1),
								 const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
								 int                               sUNUSED(state_THEN_B_2),
								 SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
								 SpecifiedStateIdentifiers_vector &sUNUSED(spec_Identifiers_THEN_B_3))
    {
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	int N_THEN_3_Bits = 0;
	int N_Equalities = spec_Identifiers_THEN_A_3.size();
	
	for (int equality = 0; equality < N_Equalities; ++equality)
	{
	    N_THEN_3_Bits += spec_Identifiers_THEN_A_3[equality].get_StateIdentifier()->get_Log2_StateCount();
	}
	aux_Variable_cnt += 1;

	return (1 + N_THEN_1_Bits + N_THEN_2_Bits + 2 * N_THEN_3_Bits);
    }

	
    int sAdvancedClauseGenerator::generate_LargeConditionalEquality(FILE                             *fw,
								    const sSpecifiedStateIdentifier  &spec_identifier_IF_A,
								    int                               state_IF_B,
								    const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_1,
								    int                               state_THEN_B_1,
								    const sSpecifiedStateIdentifier  &spec_identifier_THEN_A_2,
								    int                               state_THEN_B_2,
								    SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_A_3,
								    SpecifiedStateIdentifiers_vector &spec_Identifiers_THEN_B_3,
								    bool                              string)
    {
	int Clause_cnt = 0;

	int N_IF_Bits = spec_identifier_IF_A.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_1_Bits = spec_identifier_THEN_A_1.get_StateIdentifier()->get_Log2_StateCount();
	int N_THEN_2_Bits = spec_identifier_THEN_A_2.get_StateIdentifier()->get_Log2_StateCount();

	int aux_cnf_variable = m_variable_store->get_Last_CNFVariable();
	m_variable_store->alloc_CNFVariables(1);

	for (int bit_1 = 0; bit_1 < N_IF_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_IF_B) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "%s ", spec_identifier_IF_A.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "%d ", spec_identifier_IF_A.calc_CNF(bit_1));
		}
	    }
	}
	if (string)
	{
	    fprintf(fw, "%s 0\n", "auxiliary");
	}
	else
	{
	    fprintf(fw, "%d 0\n", aux_cnf_variable);
	}
	++Clause_cnt;


	for (int bit_1 = 0; bit_1 < N_THEN_1_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_1) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_1.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_1.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	for (int bit_1 = 0; bit_1 < N_THEN_2_Bits; ++bit_1)
	{
	    if (calc_Bit(bit_1, state_THEN_B_2) == 1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s %s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d %d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    else
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s 0\n", "auxiliary", spec_identifier_THEN_A_2.calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d 0\n", aux_cnf_variable, spec_identifier_THEN_A_2.calc_CNF(bit_1));
		}
	    }
	    ++Clause_cnt;
	}

	int N_Equalities = spec_Identifiers_THEN_A_3.size();
	sASSERT(spec_Identifiers_THEN_A_3.size() == spec_Identifiers_THEN_B_3.size());
	
	for (int equality = 0; equality < N_Equalities; ++equality)
	{
	    sASSERT(spec_Identifiers_THEN_A_3[equality].get_StateIdentifier()->get_Log2_StateCount() == spec_Identifiers_THEN_B_3[equality].get_StateIdentifier()->get_Log2_StateCount());
	    int N_THEN_3_Bits = spec_Identifiers_THEN_A_3[equality].get_StateIdentifier()->get_Log2_StateCount();
	    
	    for (int bit_1 = 0; bit_1 < N_THEN_3_Bits; ++bit_1)
	    {
		if (string)
		{
		    fprintf(fw, "-%s -%s %s 0\n", "auxiliary", spec_Identifiers_THEN_A_3[equality].calc_String(bit_1).c_str(), spec_Identifiers_THEN_B_3[equality].calc_String(bit_1).c_str());
		    fprintf(fw, "-%s %s -%s 0\n", "auxiliary", spec_Identifiers_THEN_A_3[equality].calc_String(bit_1).c_str(), spec_Identifiers_THEN_B_3[equality].calc_String(bit_1).c_str());
		}
		else
		{
		    fprintf(fw, "-%d -%d %d 0\n", aux_cnf_variable, spec_Identifiers_THEN_A_3[equality].calc_CNF(bit_1), spec_Identifiers_THEN_B_3[equality].calc_CNF(bit_1));
		    fprintf(fw, "-%d %d -%d 0\n", aux_cnf_variable, spec_Identifiers_THEN_A_3[equality].calc_CNF(bit_1), spec_Identifiers_THEN_B_3[equality].calc_CNF(bit_1));
		}
		Clause_cnt += 2;
	    }
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.get_CurrentPhase().m_produced_cnf_Clauses += Clause_cnt;
	}
        #endif

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::count_AllDifferenceConstraint(int                              &aux_Variable_cnt,
								SpecifiedStateIdentifiers_vector &spec_Identifiers) const
    {
	int Clause_cnt = 0;

	int N_Identifiers = spec_Identifiers.size();
	int N_Bits = spec_Identifiers[0].get_StateIdentifier()->get_Log2_StateCount();
	int N_States = spec_Identifiers[0].get_StateIdentifier()->get_StateCount();
	int log2_Identifiers = sIndexableStateIdentifier::calc_Log2(N_Identifiers);

	aux_Variable_cnt += 2 * N_Identifiers * (log2_Identifiers) + N_Identifiers * N_Bits + N_Identifiers * N_Identifiers * 2 + N_Bits * (N_Identifiers - 1);

	Clause_cnt += 2 * N_Identifiers * (sIndexableStateIdentifier::calc_Exp2(log2_Identifiers) - N_Identifiers) + N_Identifiers * (sIndexableStateIdentifier::calc_Exp2(N_Bits) - N_States);
	Clause_cnt += (N_Identifiers * N_Identifiers * ((1 + log2_Identifiers + 2 * N_Bits) + (1 + log2_Identifiers))) + (1 + (N_Bits * (N_Bits + 1))) * (N_Identifiers - 1);

	return Clause_cnt;
    }


    int sAdvancedClauseGenerator::generate_AllDifferenceConstraint(FILE                             *fw,
								   SpecifiedStateIdentifiers_vector &spec_Identifiers,
								   bool                              string)
    {
	int Clause_cnt = 0;
	int N_Identifiers = spec_Identifiers.size();
	int N_States = spec_Identifiers[0].get_StateIdentifier()->get_StateCount();

	sIndexableStateIdentifier bijection_A_auxiliary(m_variable_store, "bijection_A-" + sInt_32_to_String(m_aux_Identifier_cnt++), N_Identifiers, sIntegerScope(0, N_Identifiers - 1));
	sIndexableStateIdentifier bijection_B_auxiliary(m_variable_store, "bijection_B-" + sInt_32_to_String(m_aux_Identifier_cnt++), N_Identifiers, sIntegerScope(0, N_Identifiers - 1));
	sIndexableStateIdentifier ordered_auxiliary(m_variable_store, "ordered-" + sInt_32_to_String(m_aux_Identifier_cnt++), N_States, sIntegerScope(0, N_Identifiers - 1));

	for (int identifier = 0; identifier < N_Identifiers; ++identifier)
	{
	    generate_Alignment(fw, sSpecifiedStateIdentifier(&bijection_A_auxiliary, sIntegerIndex(identifier)));
	    generate_Alignment(fw, sSpecifiedStateIdentifier(&bijection_B_auxiliary, sIntegerIndex(identifier)));
	    generate_Alignment(fw, sSpecifiedStateIdentifier(&ordered_auxiliary, sIntegerIndex(identifier)));
	}

	for (int identifier_A = 0; identifier_A < N_Identifiers; ++identifier_A)
	{
	    for (int identifier_B = 0; identifier_B < N_Identifiers; ++identifier_B)
	    {
		Clause_cnt += generate_ConditionalEquality(fw,
							   sSpecifiedStateIdentifier(&bijection_A_auxiliary, sIntegerIndex(identifier_A)),
							   identifier_B,
							   sSpecifiedStateIdentifier(&bijection_B_auxiliary, sIntegerIndex(identifier_B)),
							   identifier_A,
							   spec_Identifiers[identifier_A],
							   sSpecifiedStateIdentifier(&ordered_auxiliary, sIntegerIndex(identifier_B)),
							   string);
		Clause_cnt += generate_ConditionalEquality(fw,
							   sSpecifiedStateIdentifier(&bijection_B_auxiliary, sIntegerIndex(identifier_B)),
							   identifier_A,
							   sSpecifiedStateIdentifier(&bijection_A_auxiliary, sIntegerIndex(identifier_A)),
							   identifier_B,							   
							   string);
	    }
	}

	for (int ordered = 0; ordered < N_Identifiers - 1; ++ordered)
	{
	    Clause_cnt += generate_LEXLess_Constraint(fw,
						      sSpecifiedStateIdentifier(&ordered_auxiliary, sIntegerIndex(ordered)),
						      sSpecifiedStateIdentifier(&ordered_auxiliary, sIntegerIndex(ordered + 1)),
						      string);
	}

	return Clause_cnt;
    }


/*----------------------------------------------------------------------------*/

} // namespace sReloc
