/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                             reLOC 0.06-hanshin                             */
/*                                                                            */
/*                   (C) Copyright 2011-2013 Pavel Surynek                    */
/*            http://www.surynek.com | <pavel.surynek@mff.cuni.cz>            */
/*                                                                            */
/*                                                                            */
/*============================================================================*/
// compress.cpp / 0.06-hanshin_029
/*----------------------------------------------------------------------------*/
//
// Compression tools for relocation problem solutions.
//
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "compile.h"
#include "defs.h"
#include "result.h"
#include "multirobot.h"
#include "compress.h"
#include "statistics.h"

using namespace sReloc;


/*----------------------------------------------------------------------------*/

namespace sReloc
{


/*----------------------------------------------------------------------------*/
// sMultirobotSolutionCompressor

    const sString sMultirobotSolutionCompressor::CNF_INVERSE_FILENAME_PREFIX = "compress/makespan_compression_input.inv";
    const sString sMultirobotSolutionCompressor::CNF_ADVANCED_FILENAME_PREFIX = "compress/makespan_compression_input.adv";
    const sString sMultirobotSolutionCompressor::CNF_DIFFERENTIAL_FILENAME_PREFIX = "compress/makespan_compression_input.dif";
    const sString sMultirobotSolutionCompressor::CNF_BIJECTION_FILENAME_PREFIX = "compress/makespan_compression_input.bij";
    const sString sMultirobotSolutionCompressor::CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX = "compress/makespan_compression_input.hadv";
    const sString sMultirobotSolutionCompressor::CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX = "compress/makespan_compression_input.hdif";
    const sString sMultirobotSolutionCompressor::CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX = "compress/makespan_compression_input.hbij";

    const sString sMultirobotSolutionCompressor::OUTPUT_INVERSE_FILENAME_PREFIX = "compress/makespan_compression_output.inv";
    const sString sMultirobotSolutionCompressor::OUTPUT_ADVANCED_FILENAME_PREFIX = "compress/makespan_compression_output.adv";
    const sString sMultirobotSolutionCompressor::OUTPUT_DIFFERENTIAL_FILENAME_PREFIX = "compress/makespan_compression_output.dif";
    const sString sMultirobotSolutionCompressor::OUTPUT_BIJECTION_FILENAME_PREFIX = "compress/makespan_compression_output.bij";
    const sString sMultirobotSolutionCompressor::OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX = "compress/makespan_compression_output.hadv";
    const sString sMultirobotSolutionCompressor::OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX = "compress/makespan_compression_output.hdif";
    const sString sMultirobotSolutionCompressor::OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX = "compress/makespan_compression_output.hbij";


/*----------------------------------------------------------------------------*/

    sMultirobotSolutionCompressor::CompressionRecord::CompressionRecord()
	: m_makespan(MAKESPAN_UNDEFINED)
    {
	// nothing
    }


    sMultirobotSolutionCompressor::CompressionRecord::CompressionRecord(int start_step, int final_step)
	: m_makespan(MAKESPAN_UNDEFINED)
	, m_start_step(start_step)
	, m_final_step(final_step)
    {
	// nothing
    }


    sMultirobotSolutionCompressor::CompressionRecord::CompressionRecord(int                        makespan,
									int                        start_step,
									int                        final_step,
									CompressionRecord         *left_component,
									CompressionRecord         *right_component,
									const sMultirobotSolution &compressed_sub_solution)
	: m_makespan(makespan)
	, m_start_step(start_step)
	, m_final_step(final_step)
	, m_left_component(left_component)
	, m_right_component(right_component)
	, m_compressed_sub_solution(compressed_sub_solution)
    {
	// nothing
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolutionCompressor::ProcessingArgument_IN::ProcessingArgument_IN()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument_IN::ProcessingArgument_IN(const sRobotArrangement   &initial_arrangement,
										const sMultirobotSolution &original_solution,
										const sUndirectedGraph    &environment)
	: m_initial_arrangement(initial_arrangement)
	, m_original_solution(original_solution)
	, m_environment(&environment)
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument_OUT::ProcessingArgument_OUT()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument_CTX::ProcessingArgument_CTX()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument_CTX::ProcessingArgument_CTX(sMultirobotSolutionCompressor &compressor)
	: m_compressor(&compressor)
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument::ProcessingArgument()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessingArgument::ProcessingArgument(const ProcessingArgument_CTX &ctx_arg,
									  const ProcessingArgument_IN  &in_arg)
	: m_ctx_arg(ctx_arg)
	, m_in_arg(in_arg)
    {
	// nothing
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolutionCompressor::ProcessedSolutionRecord::ProcessedSolutionRecord()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::ProcessedSolutionRecord::ProcessedSolutionRecord(int start_step, int final_step, const sMultirobotSolution &processed_solution)
	: m_start_step(start_step)
	, m_final_step(final_step)
	, m_processed_solution(processed_solution)
    {
	// nothing
    }

    bool sMultirobotSolutionCompressor::ProcessedSolutionRecord::operator==(const ProcessedSolutionRecord &processed_record) const
    {
	return (m_start_step == processed_record.m_start_step);
    }


    bool sMultirobotSolutionCompressor::ProcessedSolutionRecord::operator<(const ProcessedSolutionRecord &processed_record) const
    {
	return (m_start_step < processed_record.m_start_step);
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolutionCompressor::AttemptDatabaseRecord::AttemptDatabaseRecord()
    {
	// nothing
    }


    sMultirobotSolutionCompressor::AttemptDatabaseRecord::AttemptDatabaseRecord(const sRobotArrangement &initial_arrangement, const sRobotArrangement &goal_arrangement)
	: m_initial_arrangement(initial_arrangement)
	, m_goal_arrangement(goal_arrangement)
    {
	// nothing
    }

    
    bool sMultirobotSolutionCompressor::AttemptDatabaseRecord::operator==(const AttemptDatabaseRecord &attempt_record) const
    {
	return (   m_initial_arrangement == attempt_record.m_initial_arrangement
		&& m_goal_arrangement == attempt_record.m_goal_arrangement);
    }


    bool sMultirobotSolutionCompressor::AttemptDatabaseRecord::operator<(const AttemptDatabaseRecord &attempt_record) const
    {
	return (    m_initial_arrangement < attempt_record.m_initial_arrangement
		    || (m_initial_arrangement == attempt_record.m_initial_arrangement && m_goal_arrangement < attempt_record.m_goal_arrangement));
    }


/*----------------------------------------------------------------------------*/

    sMultirobotSolutionCompressor::sMultirobotSolutionCompressor(const sString &minisat_path,
								 int            minisat_timeout,
								 int            total_timeout,
								 int            makespan_upper_bound,
								 int            N_parallel_Threads,
								 Encoding       encoding)
	: m_minisat_timeout(minisat_timeout)
	, m_total_timeout(total_timeout)
	, m_makespan_upper_bound(makespan_upper_bound)
	, m_minisat_path(minisat_path)
	, m_N_parallel_Threads(N_parallel_Threads)
	, m_encoding(encoding)
    {
	// nothing
    }


    sMultirobotSolutionCompressor::sMultirobotSolutionCompressor(const sMultirobotSolutionCompressor &solution_compressor)
	: m_minisat_timeout(solution_compressor.m_minisat_timeout)
	, m_total_timeout(solution_compressor.m_total_timeout)
	, m_makespan_upper_bound(solution_compressor.m_makespan_upper_bound)
	, m_minisat_path(solution_compressor.m_minisat_path)
	, m_N_parallel_Threads(solution_compressor.m_N_parallel_Threads)
	, m_encoding(solution_compressor.m_encoding)
    {
	// nothing
    }


    const sMultirobotSolutionCompressor& sMultirobotSolutionCompressor::operator=(const sMultirobotSolutionCompressor &solution_compressor)
    {
	m_minisat_timeout = solution_compressor.m_minisat_timeout;
	m_total_timeout = solution_compressor.m_total_timeout;
	m_makespan_upper_bound = solution_compressor.m_makespan_upper_bound;
	m_minisat_path = solution_compressor.m_minisat_path;
        m_N_parallel_Threads = solution_compressor.m_N_parallel_Threads;
	m_encoding = solution_compressor.m_encoding;

	return *this;
    }


    int sMultirobotSolutionCompressor::calc_MakespanLowerBound(const sRobotArrangement                     &start_arrangement,
							       const sRobotArrangement                     &final_arrangement,
							       const sUndirectedGraph::Distances_2d_vector &all_pairs_Distances)
    {
	int maximum_distance = 0;
	int N_Robots = start_arrangement.get_RobotCount();

	for (int robot_id = 1; robot_id <= N_Robots; ++robot_id)
	{
	    int u_id = start_arrangement.get_RobotLocation(robot_id);
	    int v_id = final_arrangement.get_RobotLocation(robot_id);

	    int distance = all_pairs_Distances[u_id][v_id];

	    if (distance > maximum_distance)
	    {
		maximum_distance = distance;
	    }
	}

	return maximum_distance;
    }


/*----------------------------------------------------------------------------*/

    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan(const sRobotArrangement &start_arrangement,
								   const sRobotArrangement &final_arrangement,
								   const sUndirectedGraph  &environment,
								   int                      makespan_upper_bound,
								   int                     &optimal_makespan,
								   int                      thread_id)
    {
	sMultirobotEncodingContext_CNFsat dummy_final_encoding_context;

	return compute_OptimalMakespan(start_arrangement, final_arrangement, environment, makespan_upper_bound, optimal_makespan, dummy_final_encoding_context, thread_id);
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan_(const sRobotArrangement &start_arrangement,
								    const sRobotArrangement &final_arrangement,
								    const sUndirectedGraph  &environment,
								    int                      makespan_upper_bound,
								    int                     &optimal_makespan,
								    int                      thread_id)
    {
	sMultirobotEncodingContext_CNFsat dummy_final_encoding_context;

	return compute_OptimalMakespan_(start_arrangement, final_arrangement, environment, makespan_upper_bound, optimal_makespan, dummy_final_encoding_context, thread_id);
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan(const sRobotArrangement &start_arrangement,
								   const sRobotArrangement &final_arrangement,
								   const sUndirectedGraph  &environment,
								   int                      makespan_lower_bound,
								   int                      makespan_upper_bound,
								   int                     &optimal_makespan,
								   int                      thread_id)
    {
	sMultirobotEncodingContext_CNFsat dummy_final_encoding_context;

	return compute_OptimalMakespan(start_arrangement, final_arrangement, environment, makespan_lower_bound, makespan_upper_bound, optimal_makespan, dummy_final_encoding_context, thread_id);
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan(const sRobotArrangement           &start_arrangement,
								   const sRobotArrangement           &final_arrangement,
								   const sUndirectedGraph            &environment,
								   int                                makespan_upper_bound,
								   int                               &optimal_makespan,
								   sMultirobotEncodingContext_CNFsat &final_encoding_context,
								   int                                thread_id)
    {
	sResult result;
	sMultirobotInstance instance(environment, start_arrangement, final_arrangement);

	int N_Layers = makespan_upper_bound + 1;
	optimal_makespan = MAKESPAN_UNDEFINED;

	while (N_Layers >= 2)
	{
	    #ifdef sVERBOSE
	    printf("Solving layer: %d\n", N_Layers);
	    #endif

	    sMultirobotEncodingContext_CNFsat encoding_context(N_Layers);
	    sString cnf_filename, output_filename;

	    switch (m_encoding)
	    {
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicAdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicDifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicBijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_DifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_BijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_AdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_INVERSE:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_InverseCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }

	    sString system_call;
	    if (m_minisat_timeout != MINISAT_TIMEOUT_UNDEFINED)
	    {
		system_call = m_minisat_path + " -cpu-lim=" + sInt_32_to_String(m_minisat_timeout) + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    else
	    {
		system_call = m_minisat_path + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    int system_result = system(system_call.c_str());

	    if (system_result < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_SYSTEM_CALL_ERROR;
	    }
            #ifdef sSTATISTICS
	    {
		++s_GlobalPhaseStatistics.get_CurrentPhase().m_total_sat_solver_Calls;
	    }
	    #endif

	    FILE *fr;
	    if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	    }
	    char answer[32];
	    fscanf(fr, "%s\n", answer);
	    fclose(fr);

            #ifndef sDEBUG
	    {
		if (unlink(cnf_filename.c_str()) < 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		}
	    }
            #endif

	    if (strcmp(answer, "UNSAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_UNSAT_sat_solver_Calls;
		}
                #endif
		return sRESULT_SUCCESS;
	    }
	    else if (strcmp(answer, "SAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_SAT_sat_solver_Calls;
		}
                #endif
		optimal_makespan = N_Layers - 1;
		final_encoding_context = encoding_context;
	    }
	    else if (strcmp(answer, "INDET") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_INDET_sat_solver_Calls;
		}
                #endif
		optimal_makespan = MAKESPAN_UNDEFINED;
		return sRESULT_SUCCESS;
	    }
	    else
	    {
		sASSERT(false);
	    }
	    --N_Layers;
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan_(const sRobotArrangement           &start_arrangement,
								    const sRobotArrangement           &final_arrangement,
								    const sUndirectedGraph            &environment,
								    int                                makespan_upper_bound,
								    int                               &optimal_makespan,
								    sMultirobotEncodingContext_CNFsat &final_encoding_context,
								    int                                thread_id)
    {
	sResult result;
	sMultirobotInstance instance(environment, start_arrangement, final_arrangement);

	int N_Layers = 2;
	optimal_makespan = MAKESPAN_UNDEFINED;

	while (N_Layers <= makespan_upper_bound + 1)
	{

	    #ifdef sVERBOSE
	    printf("Solving layer: %d\n", N_Layers);
	    #endif

	    sMultirobotEncodingContext_CNFsat encoding_context(N_Layers);
	    sString cnf_filename, output_filename;

	    switch (m_encoding)
	    {
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicAdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicDifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicBijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }

	    case ENCODING_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_DifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_BijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_AdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_INVERSE:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_InverseCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }

	    sString system_call;
	    if (m_minisat_timeout != MINISAT_TIMEOUT_UNDEFINED)
	    {
		system_call = m_minisat_path + " -cpu-lim=" + sInt_32_to_String(m_minisat_timeout) + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    else
	    {
		system_call = m_minisat_path + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    int system_result = system(system_call.c_str());

	    if (system_result < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_SYSTEM_CALL_ERROR;
	    }
            #ifdef sSTATISTICS
	    {
		++s_GlobalPhaseStatistics.get_CurrentPhase().m_total_sat_solver_Calls;
	    }
	    #endif

	    FILE *fr;
	    if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	    }
	    char answer[32];
	    fscanf(fr, "%s\n", answer);
	    fclose(fr);

            #ifndef sDEBUG
	    {
		if (unlink(cnf_filename.c_str()) < 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		}
	    }
            #endif

	    if (strcmp(answer, "UNSAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_UNSAT_sat_solver_Calls;
		}
                #endif
	    }
	    else if (strcmp(answer, "SAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_SAT_sat_solver_Calls;
		}
                #endif
		optimal_makespan = N_Layers - 1;
		final_encoding_context = encoding_context;
		return sRESULT_SUCCESS;
	    }
	    else if (strcmp(answer, "INDET") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_INDET_sat_solver_Calls;
		}
                #endif
		optimal_makespan = MAKESPAN_UNDEFINED;
		return sRESULT_SUCCESS;
	    }
	    else
	    {
		sASSERT(false);
	    }
	    ++N_Layers;
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalMakespan(const sRobotArrangement           &start_arrangement,
								   const sRobotArrangement           &final_arrangement,
								   const sUndirectedGraph            &environment,
								   int                                makespan_lower_bound,
								   int                                makespan_upper_bound,
								   int                               &optimal_makespan,
								   sMultirobotEncodingContext_CNFsat &final_encoding_context,
								   int                                thread_id)
    {
	sResult result;
	sMultirobotInstance instance(environment, start_arrangement, final_arrangement);

	optimal_makespan = MAKESPAN_UNDEFINED;

	int first_unsatisfiable_makespan = makespan_lower_bound - 1;
	int last_satisfiable_makespan = makespan_upper_bound + 1;

	while (last_satisfiable_makespan - first_unsatisfiable_makespan > 1)
	{
	    int makespan_try = last_satisfiable_makespan - (last_satisfiable_makespan - first_unsatisfiable_makespan) / 2;
	    int N_Layers = makespan_try + 1;

	    #ifdef sVERBOSE
	    printf("Solving layer: %d\n", N_Layers);
	    #endif

	    sMultirobotEncodingContext_CNFsat encoding_context(N_Layers);
	    sString cnf_filename, output_filename;

	    switch (m_encoding)
	    {
	    case ENCODING_INVERSE:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_InverseCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_AdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_DifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_BijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicAdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicDifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicBijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }

	    default:
	    {
		sASSERT(false);
	    }
	    }

	    sString system_call;
	    if (m_minisat_timeout != MINISAT_TIMEOUT_UNDEFINED)
	    {
		system_call = m_minisat_path + " -cpu-lim=" + sInt_32_to_String(m_minisat_timeout) + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    else
	    {
		system_call = m_minisat_path + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    int system_result = system(system_call.c_str());

	    if (system_result < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_SYSTEM_CALL_ERROR;
	    }
            #ifdef sSTATISTICS
	    {
		++s_GlobalPhaseStatistics.get_CurrentPhase().m_total_sat_solver_Calls;
	    }
	    #endif

	    FILE *fr;
	    if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	    }

	    char answer[32];
	    fscanf(fr, "%s\n", answer);
	    fclose(fr);

            #ifndef sDEBUG
	    {
		if (unlink(cnf_filename.c_str()) < 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		}
	    }
            #endif

	    if (strcmp(answer, "UNSAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_UNSAT_sat_solver_Calls;
		}
                #endif
		first_unsatisfiable_makespan = makespan_try;

                #ifndef sDEBUG
		{
		    if (unlink(output_filename.c_str()) < 0)
		    {
			return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		    }
		}
                #endif
	    }
	    else if (strcmp(answer, "SAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_SAT_sat_solver_Calls;
		}
                #endif
		final_encoding_context = encoding_context;
		last_satisfiable_makespan = makespan_try;
	    }
	    else if (strcmp(answer, "INDET") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_INDET_sat_solver_Calls;
		}
                #endif
		optimal_makespan = MAKESPAN_UNDEFINED;

                #ifndef sDEBUG
		{
		    if (unlink(output_filename.c_str()) < 0)
		    {
			return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		    }
		}
                #endif
		return sRESULT_SUCCESS;
	    }
	    else
	    {
		sASSERT(false);
	    }
	}

	if (last_satisfiable_makespan <= makespan_upper_bound)
	{
	    optimal_makespan = last_satisfiable_makespan;
	}
	else
	{
	    optimal_makespan = MAKESPAN_UNDEFINED;
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalSolution(const sRobotArrangement &start_arrangement,
								   const sRobotArrangement &final_arrangement,
								   const sUndirectedGraph  &environment,
								   int                      makespan_upper_bound,
								   int                     &optimal_makespan,
								   sMultirobotSolution     &optimal_solution,
								   int                      thread_id)
    {
	sResult result;

	sMultirobotEncodingContext_CNFsat final_encoding_context;

	result = compute_OptimalMakespan(start_arrangement, final_arrangement, environment, makespan_upper_bound, optimal_makespan, final_encoding_context, thread_id);
	if (sFAILED(result))
	{
	    return result;
	}

	if (optimal_makespan != MAKESPAN_UNDEFINED)
	{
	    switch (m_encoding)
	    {
	    case ENCODING_DIFFERENTIAL:
	    {
		result = extract_ComputedDifferentialSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_INVERSE:
	    {
		result = extract_ComputedInverseSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		result = extract_ComputedAdvancedSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		result = extract_ComputedBijectionSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalSolution_(const sRobotArrangement &start_arrangement,
								    const sRobotArrangement &final_arrangement,
								    const sUndirectedGraph  &environment,
								    int                      makespan_upper_bound,
								    int                     &optimal_makespan,
								    sMultirobotSolution     &optimal_solution,
								    int                      thread_id)
    {
	sResult result;

	sMultirobotEncodingContext_CNFsat final_encoding_context;

	result = compute_OptimalMakespan_(start_arrangement, final_arrangement, environment, makespan_upper_bound, optimal_makespan, final_encoding_context, thread_id);
	if (sFAILED(result))
	{
	    return result;
	}

	if (optimal_makespan != MAKESPAN_UNDEFINED)
	{
	    switch (m_encoding)
	    {
	    case ENCODING_DIFFERENTIAL:
	    {
		result = extract_ComputedDifferentialSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_INVERSE:
	    {
		result = extract_ComputedInverseSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		result = extract_ComputedAdvancedSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		result = extract_ComputedBijectionSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		result = extract_ComputedHeuristicAdvancedSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		result = extract_ComputedHeuristicBijectionSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		result = extract_ComputedHeuristicDifferentialSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_OptimalSolution(const sRobotArrangement &start_arrangement,
								   const sRobotArrangement &final_arrangement,
								   const sUndirectedGraph  &environment,
								   int                      makespan_lower_bound,
								   int                      makespan_upper_bound,
								   int                     &optimal_makespan,
								   sMultirobotSolution     &optimal_solution,
								   int                      thread_id)
    {
	sResult result;

	sMultirobotEncodingContext_CNFsat final_encoding_context;

	result = compute_OptimalMakespan(start_arrangement, final_arrangement, environment, makespan_lower_bound, makespan_upper_bound, optimal_makespan, final_encoding_context, thread_id);
	if (sFAILED(result))
	{
	    return result;
	}

	if (optimal_makespan != MAKESPAN_UNDEFINED)
	{	
	    switch (m_encoding)
	    {
	    case ENCODING_INVERSE:
	    {
		result = extract_ComputedInverseSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		result = extract_ComputedAdvancedSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		result = extract_ComputedBijectionSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_DIFFERENTIAL:
	    {
		result = extract_ComputedDifferentialSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		result = extract_ComputedHeuristicAdvancedSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		result = extract_ComputedHeuristicBijectionSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		result = extract_ComputedHeuristicDifferentialSolution(start_arrangement, environment, optimal_makespan, final_encoding_context, optimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }
	}

	return sRESULT_SUCCESS;
    }


/*----------------------------------------------------------------------------*/

    sResult sMultirobotSolutionCompressor::compute_SuboptimalMakespan(const sRobotArrangement &start_arrangement,
								      const sRobotArrangement &final_arrangement,
								      const sUndirectedGraph  &environment,
								      int                      makespan_lower_bound,
								      int                      makespan_upper_bound,
								      int                     &suboptimal_makespan,
								      int                      thread_id)
    {
	sMultirobotEncodingContext_CNFsat dummy_final_encoding_context;

	return compute_SuboptimalMakespan(start_arrangement, final_arrangement, environment, makespan_lower_bound, makespan_upper_bound, suboptimal_makespan, dummy_final_encoding_context, thread_id);
    }


    sResult sMultirobotSolutionCompressor::compute_SuboptimalMakespan(const sRobotArrangement           &start_arrangement,
								      const sRobotArrangement           &final_arrangement,
								      const sUndirectedGraph            &environment,
								      int                                makespan_lower_bound,
								      int                                makespan_upper_bound,
								      int                               &suboptimal_makespan,
								      sMultirobotEncodingContext_CNFsat &final_encoding_context,
								      int                                thread_id)
    {
	sResult result;
	sMultirobotInstance instance(environment, start_arrangement, final_arrangement);

	suboptimal_makespan = MAKESPAN_UNDEFINED;

	int first_unsatisfiable_makespan = makespan_lower_bound - 1;
	int last_satisfiable_makespan = makespan_upper_bound + 1;

	while (last_satisfiable_makespan - first_unsatisfiable_makespan > 1)
	{
	    int makespan_try = last_satisfiable_makespan - (last_satisfiable_makespan - first_unsatisfiable_makespan) / 2;
	    int N_Layers = makespan_try + 1;

	    sMultirobotEncodingContext_CNFsat encoding_context(N_Layers);
	    sString cnf_filename, output_filename;

	    switch (m_encoding)
	    {
	    case ENCODING_INVERSE:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_InverseCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_AdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_DifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_BijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicAdvancedCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicDifferentialCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
		}
		else
		{
		    cnf_filename = CNF_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".cnf";
		    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(N_Layers) + "-" + sInt_32_to_String(getpid()) + ".txt";
		}

		result = instance.to_File_HeuristicBijectionCNFsat(cnf_filename, encoding_context, "", false);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }

	    default:
	    {
		sASSERT(false);
	    }
	    }

	    sString system_call;
	    if (m_minisat_timeout != MINISAT_TIMEOUT_UNDEFINED)
	    {
		system_call = m_minisat_path + " -cpu-lim=" + sInt_32_to_String(m_minisat_timeout) + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    else
	    {
		system_call = m_minisat_path + " " + cnf_filename + " " + output_filename +  " 1>/dev/null";
	    }
	    int system_result = system(system_call.c_str());

	    if (system_result < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_SYSTEM_CALL_ERROR;
	    }
            #ifdef sSTATISTICS
	    {
		++s_GlobalPhaseStatistics.get_CurrentPhase().m_total_sat_solver_Calls;
	    }
	    #endif

	    FILE *fr;
	    if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	    }

	    char answer[32];
	    fscanf(fr, "%s\n", answer);
	    fclose(fr);

            #ifndef sDEBUG
	    {
		if (unlink(cnf_filename.c_str()) < 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		}
	
	    }
            #endif

	    if (strcmp(answer, "UNSAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_UNSAT_sat_solver_Calls;
		}
                #endif
		first_unsatisfiable_makespan = makespan_try;

                #ifndef sDEBUG
		{
		    if (unlink(output_filename.c_str()) < 0)
		    {
			return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		    }
		}
                #endif
	    }
	    else if (strcmp(answer, "SAT") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_SAT_sat_solver_Calls;
		}
                #endif
		final_encoding_context = encoding_context;
		suboptimal_makespan = makespan_try;
		return sRESULT_SUCCESS;
	    }
	    else if (strcmp(answer, "INDET") == 0)
	    {
                #ifdef sSTATISTICS
		{
		    ++s_GlobalPhaseStatistics.get_CurrentPhase().m_INDET_sat_solver_Calls;
		}
                #endif
		//		first_unsatisfiable_makespan = makespan_try;
		suboptimal_makespan = MAKESPAN_UNDEFINED;

                #ifndef sDEBUG
		{
		    if (unlink(output_filename.c_str()) < 0)
		    {
			return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
		    }
		}
                #endif
		return sRESULT_SUCCESS;
	    }
	    else
	    {
		sASSERT(false);
	    }
	}

	if (last_satisfiable_makespan <= makespan_upper_bound)
	{
	    suboptimal_makespan = last_satisfiable_makespan;
	}
	else
	{
	    suboptimal_makespan = MAKESPAN_UNDEFINED;
	}

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compute_SuboptimalSolution(const sRobotArrangement &start_arrangement,
								      const sRobotArrangement &final_arrangement,
								      const sUndirectedGraph  &environment,
								      int                      makespan_lower_bound,
								      int                      makespan_upper_bound,
								      int                     &suboptimal_makespan,
								      sMultirobotSolution     &suboptimal_solution,
								      int                      thread_id)
    {
	sResult result;

	sMultirobotEncodingContext_CNFsat final_encoding_context;

	result = compute_SuboptimalMakespan(start_arrangement, final_arrangement, environment, makespan_lower_bound, makespan_upper_bound, suboptimal_makespan, final_encoding_context, thread_id);
	if (sFAILED(result))
	{
	    return result;
	}

	if (suboptimal_makespan != MAKESPAN_UNDEFINED)
	{
	    switch (m_encoding)
	    {
	    case ENCODING_INVERSE:
	    {
		result = extract_ComputedInverseSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_ADVANCED:
	    {
		result = extract_ComputedAdvancedSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_DIFFERENTIAL:
	    {
		result = extract_ComputedDifferentialSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_BIJECTION:
	    {
		result = extract_ComputedBijectionSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_ADVANCED:
	    {
		result = extract_ComputedHeuristicAdvancedSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_DIFFERENTIAL:
	    {
		result = extract_ComputedHeuristicDifferentialSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    case ENCODING_HEURISTIC_BIJECTION:
	    {
		result = extract_ComputedHeuristicBijectionSolution(start_arrangement, environment, suboptimal_makespan, final_encoding_context, suboptimal_solution, thread_id);
		if (sFAILED(result))
		{
		    return result;
		}
		break;
	    }
	    default:
	    {
		sASSERT(false);
	    }
	    }
	}

	return sRESULT_SUCCESS;
    }


/*----------------------------------------------------------------------------*/

    sResult sMultirobotSolutionCompressor::extract_ComputedInverseSolution(const sRobotArrangement                 &start_arrangement,
									   const sUndirectedGraph                  &environment,
									   int                                      computed_makespan,
									   const sMultirobotEncodingContext_CNFsat &final_encoding_context,
									   sMultirobotSolution                     &computed_solution,
									   int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_INVERSE_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	    }
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
	
	sMultirobotSolution pre_solution;
	
	int transition_neighbor = 0;
	int transition_vertex_id = -1;
	int transition_step = -1;
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("transition_action") == 0)
		    {   
			int transition_state_bit = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());
			
			if (transition_state_bit == 0)
			{
			    if (transition_vertex_id >= 0)
			    {
				int N_Neighbors = environment.get_Vertex(transition_vertex_id)->calc_NeighborCount();
				if (transition_neighbor < N_Neighbors && transition_step < computed_makespan)
				{
				    sMultirobotSolution::Move move(-1, transition_vertex_id, environment.get_Vertex(transition_vertex_id)->calc_NeighborID(transition_neighbor));
				    pre_solution.add_Move(transition_step, move);
				}
			    }
			    transition_neighbor = 0;
			}
			transition_vertex_id = sInt_32_from_String(base_name.substr(base_name.find("-") + 1).c_str());
			transition_step = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin());
			
			transition_neighbor += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(transition_state_bit) : 0;
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	sRobotArrangement current_arrangement = start_arrangement;
	
	int N_Steps = pre_solution.m_Steps.size();
	for (int step = 0; step < N_Steps; ++step)
	{
	    const sMultirobotSolution::Moves_list &step_Moves = pre_solution.m_Steps[step].m_Moves;
	    for (sMultirobotSolution::Moves_list::const_iterator move = step_Moves.begin(); move != step_Moves.end(); ++move)
	    {
		int robot_id = current_arrangement.get_VertexOccupancy(move->m_src_vrtx_id);
		
		if (robot_id >= 1)
		{
		    computed_solution.add_Move(step, sMultirobotSolution::Move(robot_id, move->m_src_vrtx_id, move->m_dest_vrtx_id));
		    sASSERT(current_arrangement.verify_Move(robot_id, move->m_dest_vrtx_id, environment));
		    current_arrangement.move_Robot(robot_id, move->m_dest_vrtx_id);
		}
	    }
	}
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedDifferentialSolution(const sRobotArrangement                 &start_arrangement,
										const sUndirectedGraph                  &sUNUSED(environment),
										int                                      computed_makespan,
										const sMultirobotEncodingContext_CNFsat &final_encoding_context,
										sMultirobotSolution                     &computed_solution,
										int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
		
	int robot_location_vertex_id = 0;
	std::vector<int> robot_Locations;
	robot_Locations.resize(start_arrangement.get_RobotCount() + 1, sRobotArrangement::UNDEFINED_LOCATION);
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("robot_location") == 0)
		    {   
			int robot_id = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin()) + 1;
			int layer = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());
			int robot_location_bit = *specified_identifier.get_ScopeIndexes()[2] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin());
			int N_Bits = *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_End() - *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin();
			robot_location_vertex_id += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(robot_location_bit) : 0;
		       
			if (robot_location_bit >= N_Bits - 1)
			{
			    if (robot_Locations[robot_id] != sRobotArrangement::UNDEFINED_LOCATION && robot_Locations[robot_id] != robot_location_vertex_id)
			    {
				sMultirobotSolution::Move move(robot_id, robot_Locations[robot_id], robot_location_vertex_id);
				computed_solution.add_Move(layer - 1, move);
			    }
			    robot_Locations[robot_id] = robot_location_vertex_id;
			    robot_location_vertex_id = 0;
			}			
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedAdvancedSolution(const sRobotArrangement                 &start_arrangement,
									    const sUndirectedGraph                  &environment,
									    int                                      computed_makespan,
									    const sMultirobotEncodingContext_CNFsat &final_encoding_context,
									    sMultirobotSolution                     &computed_solution,
									    int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
	
	sMultirobotSolution pre_solution;
	
	int transition_neighbor = 0;
	int transition_vertex_id = -1;
	int transition_step = -1;
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("transition_action") == 0)
		    {   
			int transition_state_bit = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());

			if (transition_state_bit == 0)
			{
			    if (transition_vertex_id >= 0)
			    {
				int N_Neighbors = environment.get_Vertex(transition_vertex_id)->calc_NeighborCount();
				if (transition_neighbor < N_Neighbors && transition_step < computed_makespan)
				{
				    sMultirobotSolution::Move move(-1, transition_vertex_id, environment.get_Vertex(transition_vertex_id)->calc_NeighborID(transition_neighbor));
				    pre_solution.add_Move(transition_step, move);
				}
			    }
			    transition_neighbor = 0;
			}
			transition_vertex_id = sInt_32_from_String(base_name.substr(base_name.find("-") + 1).c_str());
			transition_step = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin());

			transition_neighbor += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(transition_state_bit) : 0;
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	sRobotArrangement current_arrangement = start_arrangement;
	
	int N_Steps = pre_solution.m_Steps.size();
	for (int step = 0; step < N_Steps; ++step)
	{
	    const sMultirobotSolution::Moves_list &step_Moves = pre_solution.m_Steps[step].m_Moves;
	    for (sMultirobotSolution::Moves_list::const_iterator move = step_Moves.begin(); move != step_Moves.end(); ++move)
	    {
		int robot_id = current_arrangement.get_VertexOccupancy(move->m_src_vrtx_id);
		
		if (robot_id >= 1)
		{
		    computed_solution.add_Move(step, sMultirobotSolution::Move(robot_id, move->m_src_vrtx_id, move->m_dest_vrtx_id));
		    sASSERT(current_arrangement.verify_Move(robot_id, move->m_dest_vrtx_id, environment));
		    current_arrangement.move_Robot(robot_id, move->m_dest_vrtx_id);
		}
	    }
	}
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedBijectionSolution(const sRobotArrangement                 &start_arrangement,
									     const sUndirectedGraph                  &sUNUSED(environment),
									     int                                      computed_makespan,
									     const sMultirobotEncodingContext_CNFsat &final_encoding_context,
									     sMultirobotSolution                     &computed_solution,
									     int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
		
	int robot_location_vertex_id = 0;
	std::vector<int> robot_Locations;
	robot_Locations.resize(start_arrangement.get_RobotCount() + 1, sRobotArrangement::UNDEFINED_LOCATION);
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("robot_location") == 0)
		    {
			int robot_id = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin()) + 1;
			int layer = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());
			int robot_location_bit = *specified_identifier.get_ScopeIndexes()[2] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin());
			int N_Bits = *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_End() - *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin();
			robot_location_vertex_id += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(robot_location_bit) : 0;
		       
			if (robot_location_bit >= N_Bits - 1)
			{
			    if (robot_Locations[robot_id] != sRobotArrangement::UNDEFINED_LOCATION && robot_Locations[robot_id] != robot_location_vertex_id)
			    {
				sMultirobotSolution::Move move(robot_id, robot_Locations[robot_id], robot_location_vertex_id);
				computed_solution.add_Move(layer - 1, move);
			    }
			    robot_Locations[robot_id] = robot_location_vertex_id;
			    robot_location_vertex_id = 0;
			}			
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedHeuristicAdvancedSolution(const sRobotArrangement                 &start_arrangement,
										     const sUndirectedGraph                  &environment,
										     int                                      computed_makespan,
										     const sMultirobotEncodingContext_CNFsat &final_encoding_context,
										     sMultirobotSolution                     &computed_solution,
										     int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_HEURISTIC_ADVANCED_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
	
	sMultirobotSolution pre_solution;
	
	int transition_neighbor = 0;
	int transition_vertex_id = -1;
	int transition_step = -1;
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("transition_action") == 0)
		    {   
			int transition_state_bit = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());

			if (transition_state_bit == 0)
			{
			    if (transition_vertex_id >= 0)
			    {
				int N_Neighbors = environment.get_Vertex(transition_vertex_id)->calc_NeighborCount();
				if (transition_neighbor < N_Neighbors && transition_step < computed_makespan)
				{
				    sMultirobotSolution::Move move(-1, transition_vertex_id, environment.get_Vertex(transition_vertex_id)->calc_NeighborID(transition_neighbor));
				    pre_solution.add_Move(transition_step, move);
				}
			    }
			    transition_neighbor = 0;
			}
			transition_vertex_id = sInt_32_from_String(base_name.substr(base_name.find("-") + 1).c_str());
			transition_step = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin());

			transition_neighbor += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(transition_state_bit) : 0;
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	sRobotArrangement current_arrangement = start_arrangement;
	
	int N_Steps = pre_solution.m_Steps.size();
	for (int step = 0; step < N_Steps; ++step)
	{
	    const sMultirobotSolution::Moves_list &step_Moves = pre_solution.m_Steps[step].m_Moves;
	    for (sMultirobotSolution::Moves_list::const_iterator move = step_Moves.begin(); move != step_Moves.end(); ++move)
	    {
		int robot_id = current_arrangement.get_VertexOccupancy(move->m_src_vrtx_id);
		
		if (robot_id >= 1)
		{
		    computed_solution.add_Move(step, sMultirobotSolution::Move(robot_id, move->m_src_vrtx_id, move->m_dest_vrtx_id));
		    sASSERT(current_arrangement.verify_Move(robot_id, move->m_dest_vrtx_id, environment));
		    current_arrangement.move_Robot(robot_id, move->m_dest_vrtx_id);
		}
	    }
	}
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedHeuristicDifferentialSolution(const sRobotArrangement                 &start_arrangement,
											 const sUndirectedGraph                  &sUNUSED(environment),
											 int                                      computed_makespan,
											 const sMultirobotEncodingContext_CNFsat &final_encoding_context,
											 sMultirobotSolution                     &computed_solution,
											 int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_HEURISTIC_DIFFERENTIAL_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
		
	int robot_location_vertex_id = 0;
	std::vector<int> robot_Locations;
	robot_Locations.resize(start_arrangement.get_RobotCount() + 1, sRobotArrangement::UNDEFINED_LOCATION);
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("robot_location") == 0)
		    {   
			int robot_id = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin()) + 1;
			int layer = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());
			int robot_location_bit = *specified_identifier.get_ScopeIndexes()[2] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin());
			int N_Bits = *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_End() - *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin();
			robot_location_vertex_id += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(robot_location_bit) : 0;
		       
			if (robot_location_bit >= N_Bits - 1)
			{
			    if (robot_Locations[robot_id] != sRobotArrangement::UNDEFINED_LOCATION && robot_Locations[robot_id] != robot_location_vertex_id)
			    {
				sMultirobotSolution::Move move(robot_id, robot_Locations[robot_id], robot_location_vertex_id);
				computed_solution.add_Move(layer - 1, move);
			    }
			    robot_Locations[robot_id] = robot_location_vertex_id;
			    robot_location_vertex_id = 0;
			}			
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::extract_ComputedHeuristicBijectionSolution(const sRobotArrangement                 &start_arrangement,
										      const sUndirectedGraph                  &sUNUSED(environment),
										      int                                      computed_makespan,
										      const sMultirobotEncodingContext_CNFsat &final_encoding_context,
										      sMultirobotSolution                     &computed_solution,
										      int                                      thread_id)
    {
	sString output_filename;
	
	if (thread_id != THREAD_ID_UNDEFINED)
	{
	    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + "#" + sInt_32_to_String(thread_id) + ".txt";
	}
	else
	{
	    output_filename = OUTPUT_HEURISTIC_BIJECTION_FILENAME_PREFIX + "_" + sInt_32_to_String(computed_makespan + 1) + "-" + sInt_32_to_String(getpid()) + ".txt";
	}

	FILE *fr;
	
	if ((fr = fopen(output_filename.c_str(), "r")) == NULL)
	{
	    return sMULTIROBOT_SOLUTION_COMPRESSOR_OPEN_ERROR;
	}
	
	char answer[32];
	fscanf(fr, "%s\n", answer);
		
	int robot_location_vertex_id = 0;
	std::vector<int> robot_Locations;
	robot_Locations.resize(start_arrangement.get_RobotCount() + 1, sRobotArrangement::UNDEFINED_LOCATION);
	
	while (!feof(fr))
	{
	    int literal;
	    fscanf(fr, "%d", &literal); 
	    
	    if (literal != 0)
	    {
		sSpecifiedIdentifier specified_identifier = final_encoding_context.translate_CNF_Variable(sABS(literal));
		
		if (!specified_identifier.is_Null())
		{
		    const sString &base_name = specified_identifier.get_IndexableIdentifier()->get_BaseName();
		    
		    if (base_name.find("robot_location") == 0)
		    {   
			int robot_id = *specified_identifier.get_ScopeIndexes()[0] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[0]->get_Begin()) + 1;
			int layer = *specified_identifier.get_ScopeIndexes()[1] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[1]->get_Begin());
			int robot_location_bit = *specified_identifier.get_ScopeIndexes()[2] - *(specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin());
			int N_Bits = *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_End() - *specified_identifier.get_IndexableIdentifier()->get_IndexScopes()[2]->get_Begin();
			robot_location_vertex_id += (literal > 0) ? sIndexableStateIdentifier::calc_Exp2(robot_location_bit) : 0;
		       
			if (robot_location_bit >= N_Bits - 1)
			{
			    if (robot_Locations[robot_id] != sRobotArrangement::UNDEFINED_LOCATION && robot_Locations[robot_id] != robot_location_vertex_id)
			    {
				sMultirobotSolution::Move move(robot_id, robot_Locations[robot_id], robot_location_vertex_id);
				computed_solution.add_Move(layer - 1, move);
			    }
			    robot_Locations[robot_id] = robot_location_vertex_id;
			    robot_location_vertex_id = 0;
			}			
		    }
		}
	    }
	    else
	    {
		break;
	    }
	}
	
	fclose(fr);

	#ifndef sDEBUG
	{
	    if (unlink(output_filename.c_str()) < 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_UNLINK_ERROR;
	    }	
	}
	#endif
	
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::shorten_Solution(const sRobotArrangement   &initial_arrangement,
							    const sMultirobotSolution &original_solution,
							    const sUndirectedGraph    &environment,
							    sMultirobotSolution       &shortened_solution,
							    int                        thread_id)
    {
        #ifdef sSTATISTICS
	{
	    if (thread_id == THREAD_ID_UNDEFINED)
	    {
		s_GlobalPhaseStatistics.enter_Phase("shortening");
	    }
	}
        #endif

	sResult result;
	Arrangements_vector unfolded_Arrangements;
	sMultirobotSolution sub_solution;

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	int original_makespan = unfolded_Arrangements.size() - 1;
	int start_step = 0;

	while (start_step < original_makespan)
	{
	    int final_step_lower_bound = start_step;
	    int final_step_upper_bound = original_makespan + 1;
	    sMultirobotSolution best_sub_solution = original_solution.extract_Subsolution(start_step, final_step_lower_bound);

	    while (final_step_upper_bound - final_step_lower_bound > 1)
	    {
		int final_step = final_step_lower_bound + (final_step_upper_bound - final_step_lower_bound) / 2;
		int makespan_lower_bound = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);

		if (makespan_lower_bound <= m_makespan_upper_bound)
		{
		    int makespan_upper_bound = sMIN(m_makespan_upper_bound, final_step - start_step + 1);
		    printf("Makespan lower:upper %d:%d\n", makespan_lower_bound, makespan_upper_bound);

		    int computed_makespan;
		    sMultirobotSolution computed_solution;

		    result = compute_OptimalSolution(unfolded_Arrangements[start_step],
						     unfolded_Arrangements[final_step],
						     environment,
						     makespan_lower_bound,
						     makespan_upper_bound,
						     computed_makespan,
						     computed_solution,
						     thread_id);
		    
		    if (sFAILED(result))
		    {
			return result;
		    }

		    if (computed_makespan != MAKESPAN_UNDEFINED)
		    {
			best_sub_solution = sMultirobotSolution(start_step, computed_solution);
			final_step_lower_bound = final_step;
		    }
		    else
		    {
			final_step_upper_bound = final_step;
		    }
		}
		else
		{
		    final_step_upper_bound = final_step;
		}
	    }

	    sub_solution = sMultirobotSolution(sub_solution, best_sub_solution);
	    start_step = final_step_lower_bound;

            #ifdef sVERBOSE
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    printf("Thread %d shortening progress = %.2f%% (current ratio = %.2f%%)\n",
			   thread_id,
			   100 * ((double)start_step / original_makespan),
			   100 * ((double)(sub_solution.get_StepCount() - sub_solution.calc_EmptySteps())  / start_step)
			);
		}
		else
		{
		    printf("Shortening progress = %.2f%% (current ratio = %.2f%%)\n",
			   100 * ((double)start_step / original_makespan),
			   100 * ((double)(sub_solution.get_StepCount() - sub_solution.calc_EmptySteps()) / start_step));
		}
	    }
            #endif
	}
	shortened_solution = sub_solution;
	shortened_solution.remove_EmptySteps();

        #ifdef sSTATISTICS
	{
	    if (thread_id == THREAD_ID_UNDEFINED)
	    {
		s_GlobalPhaseStatistics.leave_Phase();
	    }
	}
        #endif
       
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::shorten_Solution_mt(const sRobotArrangement   &initial_arrangement,
							       const sMultirobotSolution &original_solution,
							       const sUndirectedGraph    &environment,
							       sMultirobotSolution       &shortened_solution)

    {
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("shortening_MT");
	}
        #endif

	#ifdef sVERBOSE
	{
	    printf("Solution shortening(mt) commenced ...\n");
	}
	#endif

	Arrangements_vector unfolded_Arrangements;
	sMultirobotSolution sub_solution;

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	int original_makespan = unfolded_Arrangements.size() - 1;

	int N_required_Threads = sMIN(original_makespan / m_makespan_upper_bound, m_N_parallel_Threads);
	N_required_Threads = (N_required_Threads < m_N_parallel_Threads && original_makespan % m_makespan_upper_bound > 0) ? N_required_Threads + 1 : N_required_Threads;

	ProcessingArguments_vector processing_Arguments;
	processing_Arguments.resize(m_N_parallel_Threads);

	int last_finish_step = 0;
	int remaining_N_Steps = original_makespan - 1;
	
	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    processing_Arguments[thread_id].m_ctx_arg.m_compressor = this;

	    int sub_solution_N_Steps = sMIN(sMAX(m_makespan_upper_bound, remaining_N_Steps / (N_required_Threads - thread_id)), remaining_N_Steps);

	    processing_Arguments[thread_id].m_in_arg.m_original_solution = original_solution.extract_Subsolution(last_finish_step, last_finish_step + sub_solution_N_Steps);
	    processing_Arguments[thread_id].m_in_arg.m_original_solution.remove_EmptySteps();
	    processing_Arguments[thread_id].m_in_arg.m_initial_arrangement = unfolded_Arrangements[last_finish_step];
	    processing_Arguments[thread_id].m_in_arg.m_environment = &environment;
	    last_finish_step += (sub_solution_N_Steps + 1);
	    remaining_N_Steps -= (sub_solution_N_Steps + 1);

	    pthread_attr_t thread_attr;
	    if (pthread_attr_init(&thread_attr) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
	    }

	    processing_Arguments[thread_id].m_ctx_arg.m_thread_id = thread_id;

	    if (pthread_create(&processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, &thread_attr, s_shorten_Solution_mt, &processing_Arguments[thread_id]) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_CREATE_ERROR;
	    }
	    if (pthread_attr_destroy(&thread_attr) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
	    }
	}

	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    if (pthread_join(processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, NULL))
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_JOIN_ERROR;
	    }
	}

	last_finish_step = 0;

	shortened_solution = sMultirobotSolution();
	
	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    if (sFAILED(processing_Arguments[thread_id].m_out_arg.m_result))
	    {
		return processing_Arguments[thread_id].m_out_arg.m_result;
	    }
	    sMultirobotSolution sub_solution(last_finish_step, processing_Arguments[thread_id].m_out_arg.m_processed_solution);
	    last_finish_step += processing_Arguments[thread_id].m_out_arg.m_processed_solution.get_StepCount();
	    shortened_solution = sMultirobotSolution(shortened_solution, sub_solution);
	}

	#ifdef sVERBOSE
	{
	    printf("Solution shortening(mt) finished\n");
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::primeShorten_Solution(const sRobotArrangement   &initial_arrangement,
								 const sMultirobotSolution &original_solution,
								 const sUndirectedGraph    &environment,
								 sMultirobotSolution       &shortened_solution,
								 int                        thread_id)
    {
        #ifdef sSTATISTICS
	{
	    if (thread_id == THREAD_ID_UNDEFINED)
	    {
		s_GlobalPhaseStatistics.enter_Phase("prime_shortening");
	    }
	}
        #endif

	sResult result;
	Arrangements_vector unfolded_Arrangements;
	sMultirobotSolution sub_solution;

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	int original_makespan = unfolded_Arrangements.size() - 1;
	int start_step = 0;

	while (start_step < original_makespan)
	{
	    int final_step_lower_bound = start_step;
	    int final_step_upper_bound = original_makespan + 1;

	    int best_final_step = final_step_lower_bound;
	    sMultirobotSolution best_sub_solution = original_solution.extract_Subsolution(start_step, final_step_lower_bound);

	    bool optimized = false;

	    while (final_step_upper_bound - final_step_lower_bound > 1)
	    {
		int final_step = final_step_lower_bound + (final_step_upper_bound - final_step_lower_bound) / 2;
		int makespan_lower_bound = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);
		int subsolution_original_makespan = final_step - start_step;
		int makespan_upper_bound = sMIN(m_makespan_upper_bound, subsolution_original_makespan);

		if (makespan_lower_bound < makespan_upper_bound)
		{
		    int computed_makespan;
		    sMultirobotSolution computed_solution;

		    if (m_attempt_Database.find(AttemptDatabaseRecord(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step])) != m_attempt_Database.end())
		    {
			computed_makespan = MAKESPAN_UNDEFINED;
		    }
		    else
		    {
			result = compute_OptimalSolution(unfolded_Arrangements[start_step],
							 unfolded_Arrangements[final_step],
							 environment,
							 makespan_lower_bound,
							 makespan_upper_bound,
							 computed_makespan,
							 computed_solution,
							 thread_id);
		    }

                    #ifdef sDEBUG
		    {
			//printf("Computed solution:%d\n", computed_makespan);
			//computed_solution.to_Screen();

			//printf("Original solution\n");
			//original_solution.extract_Subsolution(start_step, final_step - 1).to_Screen();
		    }
                    #endif
		    
		    if (sFAILED(result))
		    {
			return result;
		    }
		    if (computed_makespan != MAKESPAN_UNDEFINED)
		    {
			if (computed_makespan < subsolution_original_makespan)
			{
			    best_sub_solution = sMultirobotSolution(start_step, computed_solution);
			    best_final_step = final_step;
			    final_step_lower_bound = final_step;
			    optimized = true;
			}
			else
			{
			    m_attempt_Database.insert(AttemptDatabaseRecord(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step]));
			    final_step_upper_bound = final_step;
			}
		    }
		    else
		    {
			final_step_upper_bound = final_step;
		    }
		}
		else
		{
		    if (makespan_lower_bound > m_makespan_upper_bound)
		    {
			final_step_upper_bound = final_step;
		    }
		    else
		    {
			final_step_lower_bound = final_step;
		    }
		}
	    }

	    if (optimized)
	    {
		sub_solution = sMultirobotSolution(sub_solution, best_sub_solution);
		//		start_step = final_step_lower_bound;
		start_step = best_final_step;
	    }
	    else
	    {
		sub_solution = sMultirobotSolution(sub_solution, original_solution.extract_Subsolution(start_step, start_step));
		start_step += 1;
	    }

	    #ifdef sDEBUG
	    {
		//printf("Start:%d Current subsolution:\n", start_step);
		//sub_solution.to_Screen();
	    }
	    #endif

            #ifdef sVERBOSE
	    {
		if (thread_id != THREAD_ID_UNDEFINED)
		{
		    printf("Thread %d shortening progress = %.2f%% (current ratio = %.2f%%)\n",
			   thread_id,
			   100 * ((double)start_step / original_makespan),
			   100 * ((double)(sub_solution.get_StepCount() - sub_solution.calc_EmptySteps())  / start_step)
			);
		}
		else
		{
		    printf("Shortening progress = %.2f%% (current ratio = %.2f%%)\n",
			   100 * ((double)start_step / original_makespan),
			   100 * ((double)(sub_solution.get_StepCount() - sub_solution.calc_EmptySteps()) / start_step));
		}
	    }
            #endif
	}
	shortened_solution = sub_solution;
	shortened_solution.remove_EmptySteps();

        #ifdef sSTATISTICS
	{
	    if (thread_id == THREAD_ID_UNDEFINED)
	    {
		s_GlobalPhaseStatistics.leave_Phase();
	    }
	}
        #endif

        #ifdef sDEBUG
	{
	    //	    printf("Shortened solution:\n");
	    //sub_solution.to_Screen();
	}
        #endif
       
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::primeShorten_Solution_mt2(const sRobotArrangement   &initial_arrangement,
								     const sMultirobotSolution &original_solution,
								     const sUndirectedGraph    &environment,
								     sMultirobotSolution       &shortened_solution)
    {
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("prime_shortening_MT");
	}
        #endif

	#ifdef sVERBOSE
	{
	    printf("Solution shortening(mt) commenced ...\n");
	}
	#endif

	Arrangements_vector unfolded_Arrangements;
	sMultirobotSolution sub_solution;

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	//	std::vector<int> Potentials;
	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	int N_Arrangements = unfolded_Arrangements.size();
	for (int i = 0; i < N_Arrangements - m_makespan_upper_bound; ++i)
	{
	    int start_step = i;
	    int final_step = start_step + m_makespan_upper_bound;

	    int dist = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);

	    printf("%d: step potential: %d,%d: --> %d <-- (%d,%d,%d)\n", i, final_step - start_step, dist, final_step - start_step - dist, start_step, final_step, dist);
	}

	int original_makespan = unfolded_Arrangements.size() - 1;

	int N_required_Threads = sMIN(original_makespan / m_makespan_upper_bound, m_N_parallel_Threads);
	N_required_Threads = (N_required_Threads < m_N_parallel_Threads && original_makespan % m_makespan_upper_bound > 0) ? N_required_Threads + 1 : N_required_Threads;

	ProcessingArguments_vector processing_Arguments;
	processing_Arguments.resize(m_N_parallel_Threads + 1);

	int last_finish_step = 0;
	int remaining_N_Steps = original_makespan - 1;
	
	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    printf("Placement: %d\n", last_finish_step);
	    processing_Arguments[thread_id].m_ctx_arg.m_compressor = this;

	    int sub_solution_N_Steps = sMIN(sMAX(m_makespan_upper_bound, remaining_N_Steps / (N_required_Threads - thread_id)), remaining_N_Steps);

	    processing_Arguments[thread_id].m_in_arg.m_original_solution = original_solution.extract_Subsolution(last_finish_step, last_finish_step + sub_solution_N_Steps);
	    processing_Arguments[thread_id].m_in_arg.m_original_solution.remove_EmptySteps();
	    processing_Arguments[thread_id].m_in_arg.m_initial_arrangement = unfolded_Arrangements[last_finish_step];
	    processing_Arguments[thread_id].m_in_arg.m_environment = &environment;
	    last_finish_step += (sub_solution_N_Steps + 1);
	    remaining_N_Steps -= (sub_solution_N_Steps + 1);

	    pthread_attr_t thread_attr;
	    if (pthread_attr_init(&thread_attr) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
	    }

	    processing_Arguments[thread_id].m_ctx_arg.m_thread_id = thread_id;

	    if (pthread_create(&processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, &thread_attr, s_prime_shorten_Solution_mt, &processing_Arguments[thread_id]) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_CREATE_ERROR;
	    }
	    if (pthread_attr_destroy(&thread_attr) != 0)
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
	    }
	}

	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    if (pthread_join(processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, NULL))
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_JOIN_ERROR;
	    }
	}

	last_finish_step = 0;

	shortened_solution = sMultirobotSolution();
	
	for (int thread_id = 0; thread_id < N_required_Threads; ++thread_id)
	{
	    if (sFAILED(processing_Arguments[thread_id].m_out_arg.m_result))
	    {
		return processing_Arguments[thread_id].m_out_arg.m_result;
	    }
	    sMultirobotSolution sub_solution(last_finish_step, processing_Arguments[thread_id].m_out_arg.m_processed_solution);
	    last_finish_step += processing_Arguments[thread_id].m_out_arg.m_processed_solution.get_StepCount();
	    shortened_solution = sMultirobotSolution(shortened_solution, sub_solution);
	}

	#ifdef sVERBOSE
	{
	    printf("Solution prime shortening(mt) finished\n");
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::primeShorten_Solution_mt(const sRobotArrangement   &initial_arrangement,
								    const sMultirobotSolution &original_solution,
								    const sUndirectedGraph    &environment,
								    sMultirobotSolution       &shortened_solution)
    {
        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("prime_shortening_MT");
	}
        #endif

	#ifdef sVERBOSE
	{
	    printf("Solution shortening(mt) commenced ...\n");
	}
	#endif

	Arrangements_vector unfolded_Arrangements;
	sMultirobotSolution sub_solution;

	original_solution.to_Screen();
	initial_arrangement.to_Screen();

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	#ifdef sDEBUG
	{
	    int N_Arrangements = unfolded_Arrangements.size();
	    for (int i = 0; i < N_Arrangements - m_makespan_upper_bound; ++i)
	    {
		int start_step = i;
		int final_step = start_step + m_makespan_upper_bound;
		
		int dist = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);

		printf("%d: step potential: %d,%d: --> %d <-- (%d,%d,%d)\n", i, final_step - start_step, dist, final_step - start_step - dist, start_step, final_step, dist);
	    }
	}
	#endif

	int original_makespan = unfolded_Arrangements.size() - 1;

	ProcessingArguments_vector processing_Arguments;
	processing_Arguments.resize(m_N_parallel_Threads);

	int last_final_step = 0;

	ProcessedSolutionRecords_set processed_solution_Records;
	int last_thread_id = 0;

	while (last_final_step < original_makespan)
	{
	    int start_step = last_final_step;
	    int final_step = start_step;

	    int distance = 0;

	    while (final_step < original_makespan)
	    {
		 ++final_step;
		 distance = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);

		 if (distance >= m_makespan_upper_bound)
		 {
		     break;
		 }
	    }

	    if (final_step - start_step > m_makespan_upper_bound)
	    {
		if (last_thread_id >= m_N_parallel_Threads)
		{
		    for (int thread_id = 0; thread_id < last_thread_id; ++thread_id)
		    {
			if (pthread_join(processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, NULL))
			{
			    return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_JOIN_ERROR;
			}
		    }
		    for (int thread_id = 0; thread_id < last_thread_id; ++thread_id)
		    {
			if (sFAILED(processing_Arguments[thread_id].m_out_arg.m_result))
			{
			    return processing_Arguments[thread_id].m_out_arg.m_result;
			}
			ProcessedSolutionRecord processed_solution_record(processing_Arguments[thread_id].m_in_arg.m_start_step,
									  processing_Arguments[thread_id].m_in_arg.m_final_step,
									  processing_Arguments[thread_id].m_out_arg.m_processed_solution);
			processed_solution_Records.insert(processed_solution_record);
		    }
		    last_thread_id = 0;
		}
		
		processing_Arguments[last_thread_id].m_ctx_arg.m_compressor = this;
		
		processing_Arguments[last_thread_id].m_in_arg.m_start_step = start_step;
		processing_Arguments[last_thread_id].m_in_arg.m_final_step = final_step;
		
		processing_Arguments[last_thread_id].m_in_arg.m_original_solution = original_solution.extract_Subsolution(start_step, final_step - 1);
		processing_Arguments[last_thread_id].m_in_arg.m_original_solution.remove_EmptySteps();
		processing_Arguments[last_thread_id].m_in_arg.m_initial_arrangement = unfolded_Arrangements[start_step];
		processing_Arguments[last_thread_id].m_in_arg.m_environment = &environment;
		
		pthread_attr_t thread_attr;
		if (pthread_attr_init(&thread_attr) != 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
		}
		processing_Arguments[last_thread_id].m_ctx_arg.m_thread_id = last_thread_id;
		
		if (pthread_create(&processing_Arguments[last_thread_id].m_ctx_arg.m_pthread_handle,
				   &thread_attr,
				   s_prime_shorten_Solution_mt,
				   &processing_Arguments[last_thread_id]) != 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_CREATE_ERROR;
		}
		
		if (pthread_attr_destroy(&thread_attr) != 0)
		{
		    return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_ATTR_ERROR;
		}
		++last_thread_id;
	    }
	    else
	    {
		final_step = last_final_step + 1;
	    }
	    last_final_step = final_step;
	}
	for (int thread_id = 0; thread_id < last_thread_id; ++thread_id)
	{
	    if (pthread_join(processing_Arguments[thread_id].m_ctx_arg.m_pthread_handle, NULL))
	    {
		return sMULTIROBOT_SOLUTION_COMPRESSOR_THREAD_JOIN_ERROR;
	    }
	}
	for (int thread_id = 0; thread_id < last_thread_id; ++thread_id)
	{
	    if (sFAILED(processing_Arguments[thread_id].m_out_arg.m_result))
	    {
		return processing_Arguments[thread_id].m_out_arg.m_result;
	    }
	    ProcessedSolutionRecord processed_solution_record(processing_Arguments[thread_id].m_in_arg.m_start_step,
							      processing_Arguments[thread_id].m_in_arg.m_final_step,
							      processing_Arguments[thread_id].m_out_arg.m_processed_solution);
	    processed_solution_Records.insert(processed_solution_record);
	}

	shortened_solution = sMultirobotSolution();
	last_final_step = 0;
	for (ProcessedSolutionRecords_set::const_iterator processed_solution =  processed_solution_Records.begin(); processed_solution !=  processed_solution_Records.end(); ++processed_solution)
	{
	    sMultirobotSolution original_part = original_solution.extract_Subsolution(last_final_step, processed_solution->m_start_step - 1);
	    sMultirobotSolution processed_part = sMultirobotSolution(processed_solution->m_start_step, processed_solution->m_processed_solution);

	    shortened_solution = sMultirobotSolution(shortened_solution, original_part);
	    shortened_solution = sMultirobotSolution(shortened_solution, processed_part);

	    last_final_step = processed_solution->m_final_step;
	}
	sMultirobotSolution final_original_part = sMultirobotSolution(last_final_step, original_solution.extract_Subsolution(last_final_step, original_makespan - 1));
	shortened_solution = sMultirobotSolution(shortened_solution, final_original_part);
	shortened_solution.remove_EmptySteps();

	#ifdef sVERBOSE
	{
	    printf("Solution prime shortening(mt) finished\n");
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::optimize_Solution(const sRobotArrangement   &initial_arrangement,
							     const sMultirobotSolution &original_solution,
							     const sUndirectedGraph    &environment,
							     sMultirobotSolution       &optimized_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution = original_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("optimization");
	}
        #endif

	#ifdef sVERBOSE
	static int pass_cnt;
	pass_cnt = 1;
	{
	    printf("Solution optimization commenced ...\n");
	}
	#endif

	double start_seconds = sGet_WC_Seconds();
	double finish_seconds = sGet_WC_Seconds();

	while (true)
	{
            #ifdef sVERBOSE
	    {
		printf("Solution shortening pass %d ...\n", pass_cnt);
		++pass_cnt;
	    }
	    #endif

	    result = shorten_Solution(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    #ifdef sVERBOSE
	    {
		printf("Solution shortening pass %d finished with ratio %.2f%%\n", pass_cnt, 100 * ((double)next_solution.get_StepCount() / previous_solution.get_StepCount()));
		printf("Intermediate statistics:\n");
		sMultirobotSolutionAnalyzer solution_analyzer;
		solution_analyzer.analyze_Solution(original_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		solution_analyzer.analyze_Solution(next_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		s_GlobalPhaseStatistics.to_Screen();
	    }
            #endif

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    optimized_solution = previous_solution;
		}
		else
		{
		    optimized_solution = next_solution;
		}
		break;
	    }
	    finish_seconds = sGet_WC_Seconds();

	    if (finish_seconds - start_seconds  > m_total_timeout)
	    {
		break;
	    }
	    previous_solution = next_solution;
	}

	#ifdef sVERBOSE
	{
	    printf("Solution optimization finished with ratio %.2f%%\n", 100 * ((double)optimized_solution.get_StepCount() / original_solution.get_StepCount()));
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::optimize_Solution_mt(const sRobotArrangement   &initial_arrangement,
								const sMultirobotSolution &original_solution,
								const sUndirectedGraph    &environment,
								sMultirobotSolution       &optimized_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution = original_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("optimization_MT");
	}
        #endif

	#ifdef sVERBOSE
	static int pass_cnt;
	pass_cnt = 0;
	{
	    printf("Solution optimization(mt) commenced ...\n");
	}
	#endif

	double start_seconds = sGet_WC_Seconds();
	double finish_seconds = sGet_WC_Seconds();

	while (true)
	{
            #ifdef sVERBOSE
	    {
		printf("Solution shortening(mt) pass %d ...\n", ++pass_cnt);
	    }
	    #endif

	    result = shorten_Solution_mt(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    #ifdef sVERBOSE
	    {
		printf("Solution shortening pass %d finished with ratio %.2f%%\n", pass_cnt, 100 * ((double)next_solution.get_StepCount() / previous_solution.get_StepCount()));
		printf("Intermediate statistics:\n");
		sMultirobotSolutionAnalyzer solution_analyzer;
		solution_analyzer.analyze_Solution(original_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		solution_analyzer.analyze_Solution(next_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		s_GlobalPhaseStatistics.to_Screen();
	    }
            #endif

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    optimized_solution = previous_solution;
		}
		else
		{
		    optimized_solution = next_solution;
		}
		break;
	    }
	    finish_seconds = sGet_WC_Seconds();

	    if (finish_seconds - start_seconds  > m_total_timeout)
	    {
		break;
	    }
	    previous_solution = next_solution;
	}

	#ifdef sVERBOSE
	{
	    printf("Solution optimization finished with ratio %.2f%%\n", 100 * ((double)optimized_solution.get_StepCount() / original_solution.get_StepCount()));
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::primeOptimize_Solution(const sRobotArrangement   &initial_arrangement,
								  const sMultirobotSolution &original_solution,
								  const sUndirectedGraph    &environment,
								  sMultirobotSolution       &optimized_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution = original_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("prime_optimization");
	}
        #endif

	#ifdef sVERBOSE
	static int pass_cnt;
	pass_cnt = 1;
	{
	    printf("Solution prime optimization commenced ...\n");
	}
	#endif

	double start_seconds = sGet_WC_Seconds();
	double finish_seconds = sGet_WC_Seconds();

	while (true)
	{
            #ifdef sVERBOSE
	    {
		printf("Solution prime shortening pass %d ...\n", pass_cnt);
		++pass_cnt;
	    }
	    #endif

	    result = primeShorten_Solution(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    #ifdef sVERBOSE
	    {
		printf("Solution prime shortening pass %d finished with ratio %.2f%%\n", pass_cnt, 100 * ((double)next_solution.get_StepCount() / previous_solution.get_StepCount()));
		printf("Intermediate statistics:\n");
		sMultirobotSolutionAnalyzer solution_analyzer;
		solution_analyzer.analyze_Solution(original_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		solution_analyzer.analyze_Solution(next_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		s_GlobalPhaseStatistics.to_Screen();
	    }
            #endif

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    optimized_solution = previous_solution;
		}
		else
		{
		    optimized_solution = next_solution;
		}
		break;
	    }
	    finish_seconds = sGet_WC_Seconds();

	    if (finish_seconds - start_seconds  > m_total_timeout)
	    {
		break;
	    }
	    previous_solution = next_solution;
	}

	#ifdef sVERBOSE
	{
	    printf("Solution prime optimization finished with ratio %.2f%%\n", 100 * ((double)optimized_solution.get_StepCount() / original_solution.get_StepCount()));
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::primeOptimize_Solution_mt(const sRobotArrangement   &initial_arrangement,
								     const sMultirobotSolution &original_solution,
								     const sUndirectedGraph    &environment,
								     sMultirobotSolution       &optimized_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution = original_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("prime_optimization_MT");
	}
        #endif

	#ifdef sVERBOSE
	static int pass_cnt;
	pass_cnt = 0;
	{
	    printf("Solution prime optimization(mt) commenced ...\n");
	}
	#endif

	double start_seconds = sGet_WC_Seconds();
	double finish_seconds = sGet_WC_Seconds();

	while (true)
	{
            #ifdef sVERBOSE
	    {
		printf("Solution prime shortening(mt) pass %d ...\n", ++pass_cnt);
	    }
	    #endif

	    result = primeShorten_Solution_mt(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    #ifdef sVERBOSE
	    {
		printf("Solution prime shortening pass %d finished with ratio %.2f%%\n", pass_cnt, 100 * ((double)next_solution.get_StepCount() / previous_solution.get_StepCount()));
		printf("Intermediate statistics:\n");
		sMultirobotSolutionAnalyzer solution_analyzer;
		solution_analyzer.analyze_Solution(original_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		solution_analyzer.analyze_Solution(next_solution, initial_arrangement, environment);
		solution_analyzer.to_Screen();
		s_GlobalPhaseStatistics.to_Screen();
	    }
            #endif

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    optimized_solution = previous_solution;
		}
		else
		{
		    optimized_solution = next_solution;
		}
		break;
	    }
	    finish_seconds = sGet_WC_Seconds();

	    if (finish_seconds - start_seconds  > m_total_timeout)
	    {
		break;
	    }

	    previous_solution = next_solution;
	}

	#ifdef sVERBOSE
	{
	    printf("Solution prime optimization finished with ratio %.2f%%\n", 100 * ((double)optimized_solution.get_StepCount() / original_solution.get_StepCount()));
	}
	#endif

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::compress_Solution(const sRobotArrangement   &initial_arrangement,
							     const sMultirobotSolution &original_solution,
							     const sUndirectedGraph    &environment,
							     sMultirobotSolution       &compressed_solution)
    {
	sResult result;
	Arrangements_vector unfolded_Arrangements;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("compression");
	}
        #endif

	sASSERT(verify_Unfolding(initial_arrangement, original_solution, environment));
	unfold_Solution(initial_arrangement, original_solution, unfolded_Arrangements);

	sUndirectedGraph::Distances_2d_vector all_pairs_Distances;
	environment.calc_AllPairsShortestPaths(all_pairs_Distances);

	CompressionMatrix_2d_vector compression_Matrix;

	int original_makespan = unfolded_Arrangements.size() - 1;
	compression_Matrix.resize(original_makespan + 1);

	for (CompressionMatrix_2d_vector::iterator matrix_row = compression_Matrix.begin(); matrix_row != compression_Matrix.end(); ++matrix_row)
	{
	    matrix_row->resize(original_makespan + 1);
	}

	for (int start_step = 0; start_step < original_makespan; ++start_step)
	{
	    int final_step = start_step + 1;
	    CompressionRecord compression_record(1, start_step, final_step, NULL, NULL, original_solution.extract_Subsolution(start_step, final_step - 1));

	    compression_Matrix[start_step][final_step] = compression_record;
	}

	for (int start_step = 0; start_step < original_makespan - 1; ++start_step)
	{
	    int final_step = start_step + 2;

	    int optimal_makespan;
	    sMultirobotSolution optimal_solution;

	    result = compute_OptimalSolution(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], environment, 3, optimal_makespan, optimal_solution);
	    if (sFAILED(result))
	    {
		return result;
	    }	    
	    CompressionRecord compression_record(optimal_makespan, start_step, final_step, NULL, NULL, sMultirobotSolution(start_step, optimal_solution));

	    compression_Matrix[start_step][final_step] = compression_record;
	}
	
	for (int makespan_distance = 3; makespan_distance <= original_makespan; ++makespan_distance)
	{
	    for (int start_step = 0; start_step < original_makespan - makespan_distance + 1; ++start_step)
	    {
		int final_step = start_step + makespan_distance;
		CompressionRecord compression_record(start_step, final_step);

		for (int left_component_span = 1; left_component_span < makespan_distance; ++left_component_span)
		{
		    int left_start_step = start_step;
		    int left_final_step = start_step + left_component_span;

		    int right_start_step = left_final_step;
		    int right_final_step = final_step;

		    int combined_makespan = compression_Matrix[left_start_step][left_final_step].m_makespan + compression_Matrix[right_start_step][right_final_step].m_makespan;

		    if (compression_record.m_makespan == MAKESPAN_UNDEFINED || combined_makespan < compression_record.m_makespan)
		    {
			compression_record.m_makespan = combined_makespan;
			
			compression_record.m_left_component = &compression_Matrix[left_start_step][left_final_step];
			compression_record.m_right_component = &compression_Matrix[right_start_step][right_final_step];
		    }
		}
		compression_record.m_compressed_sub_solution = sMultirobotSolution(compression_record.m_left_component->m_compressed_sub_solution,
										   compression_record.m_right_component->m_compressed_sub_solution);


		int makespan_lower_bound = calc_MakespanLowerBound(unfolded_Arrangements[start_step], unfolded_Arrangements[final_step], all_pairs_Distances);

		if (makespan_lower_bound <= m_makespan_upper_bound)
		{
		    int makespan_upper_bound = sMIN(m_makespan_upper_bound, final_step - start_step + 1);

		    int optimal_makespan;
		    sMultirobotSolution optimal_solution;	      

		    result = compute_OptimalSolution(unfolded_Arrangements[start_step],
						     unfolded_Arrangements[final_step],
						     environment,
						     makespan_lower_bound,
						     makespan_upper_bound,
						     optimal_makespan,
						     optimal_solution);
		    if (sFAILED(result))
		    {
			return result;
		    }
		    if (optimal_makespan != MAKESPAN_UNDEFINED && optimal_makespan < compression_record.m_makespan)
		    {
			compression_record.m_makespan = optimal_makespan;
			
			compression_record.m_left_component = NULL;
			compression_record.m_right_component = NULL;
			
			compression_record.m_compressed_sub_solution = sMultirobotSolution(start_step, optimal_solution);
		    }
		}

		compression_Matrix[start_step][final_step] = compression_record;
	    }
	}
	compressed_solution = compression_Matrix[0][original_makespan].m_compressed_sub_solution;
	compressed_solution.remove_EmptySteps();

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif
       
	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::deflate_Solution(const sRobotArrangement   &initial_arrangement,
							    const sMultirobotSolution &original_solution,
							    const sUndirectedGraph    &environment,
							    sMultirobotSolution       &deflated_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("deflation");
	}
        #endif

	result = optimize_Solution(initial_arrangement, original_solution, environment, previous_solution);

	if (sFAILED(result))
	{
	    return result;
	}
	while (true)
	{
	    result = compress_Solution(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    deflated_solution = previous_solution;
		}
		else
		{
		    deflated_solution = next_solution;
		}
		break;
	    }
	    previous_solution = next_solution;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    sResult sMultirobotSolutionCompressor::deflate_Solution_mt(const sRobotArrangement   &initial_arrangement,
							       const sMultirobotSolution &original_solution,
							       const sUndirectedGraph    &environment,
							       sMultirobotSolution       &deflated_solution)
    {
	sResult result;
	sMultirobotSolution previous_solution;
	sMultirobotSolution next_solution;

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.enter_Phase("deflation");
	}
        #endif

	result = optimize_Solution_mt(initial_arrangement, original_solution, environment, previous_solution);

	if (sFAILED(result))
	{
	    return result;
	}
	while (true)
	{
	    result = compress_Solution(initial_arrangement, previous_solution, environment, next_solution);

	    if (sFAILED(result))
	    {
		return result;
	    }

	    if (previous_solution.get_StepCount() <= next_solution.get_StepCount())
	    {
		if (previous_solution.get_MoveCount() <= next_solution.get_MoveCount())
		{
		    deflated_solution = previous_solution;
		}
		else
		{
		    deflated_solution = next_solution;
		}
		break;
	    }
	    previous_solution = next_solution;
	}

        #ifdef sSTATISTICS
	{
	    s_GlobalPhaseStatistics.leave_Phase();
	}
        #endif

	return sRESULT_SUCCESS;
    }


    void sMultirobotSolutionCompressor::unfold_Solution(const sRobotArrangement &initial_arrangement, const sMultirobotSolution &solution, Arrangements_vector &unfolded_Arrangements)
    {
	unfolded_Arrangements.push_back(initial_arrangement);
	sRobotArrangement current_arrangement = initial_arrangement;

	for (sMultirobotSolution::Steps_vector::const_iterator step =  solution.m_Steps.begin(); step != solution.m_Steps.end(); ++step)
	{
	    for (sMultirobotSolution::Moves_list::const_iterator move = step->m_Moves.begin(); move != step->m_Moves.end(); ++move)
	    {
		current_arrangement.move_Robot(move->m_robot_id, move->m_dest_vrtx_id);
	    }
	    unfolded_Arrangements.push_back(current_arrangement);
	}
	
    }


    bool sMultirobotSolutionCompressor::verify_Unfolding(const sRobotArrangement &initial_arrangement, const sMultirobotSolution &solution, const sUndirectedGraph &graph)
    {
	sRobotArrangement current_arrangement = initial_arrangement;

	for (sMultirobotSolution::Steps_vector::const_iterator step = solution.m_Steps.begin(); step != solution.m_Steps.end(); ++step)
	{
	    for (sMultirobotSolution::Moves_list::const_iterator move = step->m_Moves.begin(); move != step->m_Moves.end(); ++move)
	    {
		if (!current_arrangement.verify_Move(move->m_robot_id, move->m_dest_vrtx_id, graph))
		{
		    return false;
		}
		current_arrangement.move_Robot(move->m_robot_id, move->m_dest_vrtx_id);
	    }
	}

	return true;
    }


    void sMultirobotSolutionCompressor::to_Screen_arrangements(const Arrangements_vector &unfolded_Arrangements, const sString &indent) const
    {      
	to_Stream_arrangements(stdout, unfolded_Arrangements, indent);
    }


    void sMultirobotSolutionCompressor::to_Stream_arrangements(FILE *fw, const Arrangements_vector &unfolded_Arrangements, const sString &indent) const
    {
	fprintf(fw, "%sRobot arrangements [\n", indent.c_str());

	for (Arrangements_vector::const_iterator arrangement = unfolded_Arrangements.begin(); arrangement != unfolded_Arrangements.end(); ++arrangement)
	{
	    arrangement->to_Stream(fw, indent + sRELOC_INDENT);
	}

	fprintf(fw, "]\n");
    }


/*----------------------------------------------------------------------------*/
// Global functions

    void* s_shorten_Solution_mt(void *arg)
    {
	sMultirobotSolutionCompressor::ProcessingArgument &processing_argument = *reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(arg);

        #ifdef sVERBOSE
	{
	    printf("Thread %d started ...\n", processing_argument.m_ctx_arg.m_thread_id);
        }
        #endif

	processing_argument.m_out_arg.m_result = processing_argument.m_ctx_arg.m_compressor->shorten_Solution(processing_argument.m_in_arg.m_initial_arrangement,
													      processing_argument.m_in_arg.m_original_solution,
													      *processing_argument.m_in_arg.m_environment,
													      processing_argument.m_out_arg.m_processed_solution,
													      processing_argument.m_ctx_arg.m_thread_id);
	if (sFAILED(processing_argument.m_out_arg.m_result))
	{
	    return (reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(&processing_argument));
	}
	processing_argument.m_out_arg.m_result = sRESULT_SUCCESS;

        #ifdef sVERBOSE
	{
	    printf("Thread %d finished with ratio %.2f%%\n",
		   processing_argument.m_ctx_arg.m_thread_id,
		   100 * (double)processing_argument.m_out_arg.m_processed_solution.get_StepCount() / processing_argument.m_in_arg.m_original_solution.get_StepCount());
        }
        #endif

	return (reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(&processing_argument));
    }


    void* s_prime_shorten_Solution_mt(void *arg)
    {
	sMultirobotSolutionCompressor::ProcessingArgument &processing_argument = *reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(arg);

        #ifdef sVERBOSE
	{
	    printf("Thread %d started ...\n", processing_argument.m_ctx_arg.m_thread_id);
        }
        #endif

	processing_argument.m_out_arg.m_result = processing_argument.m_ctx_arg.m_compressor->primeShorten_Solution(processing_argument.m_in_arg.m_initial_arrangement,
														   processing_argument.m_in_arg.m_original_solution,
														   *processing_argument.m_in_arg.m_environment,
														   processing_argument.m_out_arg.m_processed_solution,
														   processing_argument.m_ctx_arg.m_thread_id);
	if (sFAILED(processing_argument.m_out_arg.m_result))
	{
	    return (reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(&processing_argument));
	}
	processing_argument.m_out_arg.m_result = sRESULT_SUCCESS;

        #ifdef sVERBOSE
	{
	    printf("Thread %d finished with ratio %.2f%%\n",
		   processing_argument.m_ctx_arg.m_thread_id,
		   100 * (double)processing_argument.m_out_arg.m_processed_solution.get_StepCount() / processing_argument.m_in_arg.m_original_solution.get_StepCount());
        }
        #endif

	return (reinterpret_cast<sMultirobotSolutionCompressor::ProcessingArgument*>(&processing_argument));
    }


/*----------------------------------------------------------------------------*/

} // namespace sReloc
